#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random
from typing import Optional

from parlai.tasks.reasoning.agents import MWPStepsReasoningTeacher
from parlai.tasks.reasoning.base import (
    AbstractQuestionAnswer,
    AbstractReason,
    AbstractReasoningTeacher,
)
from parlai.tasks.reasoning.question_answer import MultipleChoiceQuestionAnswer
from parlai.tasks.reasoning.reason_types.step_by_step import StepByStepReason

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import DialogTeacher
from parlai.utils.io import PathManager
from parlai.tasks.aqua.build import build

import os
import copy
import json


AQUA = 'AQuA'
AQUA_QUESTION_KEY = 'question'
AQUA_ANSWER_KEY = 'correct'
AQUA_OPTIONS_KEY = 'options'
AQUA_RATIONALE_KEY = 'rationale'
RATIONALE_QUESTION_TEXT = 'Can you provide a rationale for your answer?'


def _path(opt):
    build(opt)

    dt = opt['datatype'].split(':')[0]

    if dt == 'train':
        prefix = 'train'
    # Using matched set as valid and mismatched set as test
    elif dt == 'valid':
        prefix = 'dev'
    elif dt == 'test':
        prefix = 'test'
    else:
        raise RuntimeError('Not valid datatype.')

    data_path = os.path.join(opt['datapath'], AQUA, AQUA, prefix + '.tok.json')

    return data_path


def setup_data(path):
    print('loading: ' + path)

    with PathManager.open(path, 'r') as data_file:
        for line in data_file:
            question = json.loads(line)
            question_text = question[AQUA_QUESTION_KEY]
            answer = ord(question[AQUA_ANSWER_KEY]) - ord('A')
            labels = question[AQUA_OPTIONS_KEY]
            answer = [labels[answer]]
            yield (question_text, answer, None, labels), True

            # Ask for a rationale now
            rationale = [question[AQUA_RATIONALE_KEY]]
            yield (RATIONALE_QUESTION_TEXT, rationale, None, rationale), False


class DefaultTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        data_path = _path(opt)
        opt['datafile'] = data_path
        self.id = 'AQuA'

        super().__init__(opt, shared)

    def setup_data(self, path):
        return setup_data(path)


class AQuAReasoningTeacher(AbstractReasoningTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.set_defaults(
            task_prompt="Solve the following math problem. Show your work.",
            choices_prefix=["", "", "", "", ""],
        )
        return parser

    @classmethod
    def get_reason_class(self) -> AbstractReason:
        return StepByStepReason

    @classmethod
    def get_question_answer_class(self) -> AbstractQuestionAnswer:
        return MultipleChoiceQuestionAnswer

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)
        opt['datafile'] = _path(opt)
        self.id = 'aqua'
        super().__init__(opt, shared)

    def get_data_for_fold(self, fold):
        with PathManager.open(self.opt['datafile']) as f:
            data = f.readlines()

        for line in data:
            problem = json.loads(line)
            question = problem["question"]
            answer = problem["correct"]
            #  A -> 0; B -> 1; C -> 2; D -> 3; E -> 4
            answer_idx = ord(answer) - 65
            options = problem["options"]
            rationale = problem["rationale"]
            steps = rationale.split("\n")
            yield {
                "question": question,
                "steps": steps,
                "answer_idx": answer_idx,
                "answer_choices": options,
            }


class AQuAStepByStepReasoningTeacher(MWPStepsReasoningTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.set_defaults(
            task_prompt="Solve the following math problem. Show your work.",
            choices_prefix=["", "", "", "", ""],
        )
        parser.add_argument(
            "--extrinsic-step",
            type=bool,
            default=False,
        )
        return parser

    def __init__(self, opt, shared=None):
        self.datatype = opt['datatype']
        build(opt)  # NOTE: the call to build here
        opt['datafile'] = _path(opt)
        self.id = 'aqua'
        self.aqua_random = random.Random(42)
        super().__init__(opt, shared)

    def get_data_for_fold(self, fold):
        extrinsic_step = self.opt.get("extrinsic_step", False)

        with PathManager.open(self.opt['datafile']) as f:
            data = f.readlines()

        messages = []
        for line in data:
            problem = json.loads(line)
            question = problem["question"].replace(' .', '.').replace(' ,', ',')
            options = {}
            for x in problem["options"]:
                try:
                    a, b = x.split(' ) ')
                except ValueError:
                    if '(' not in x:
                        # some lines have typos and duplicate indices
                        a, c, b = x.split(' ) ')
                    else:
                        # example: B ) ( 4 ! - 1 ) * ( 5 ! - 1 ) * ( 6 ! - 1 )
                        a = x.split(' ) ')[0]
                        b = " ) ".join(x.split(' ) ')[1:])
                options[a] = b
            answer = options[problem["correct"]]
            rationale = problem["rationale"]
            steps = rationale.split("\n")
            final_steps = []
            # some steps are just dots or empty
            for step in steps:
                while len(step) > 0 and (step[-1] == '.' or step[-1] == ' '):
                    step = step[:-1]
                if len(step) > 0:
                    final_steps.append(step)
            m = {"question": question, "steps": final_steps, "answer": answer}
            if not extrinsic_step:
                yield m
            else:
                messages.append(m)

        if extrinsic_step:
            for line in messages:
                rand_steps = self.aqua_random.choice(messages)["steps"]
                random_step = self.aqua_random.choice(rand_steps)
                yield {
                    "question": line["question"],
                    "answer": line["answer"],
                    "steps": line["steps"],
                    "extrinsic_step": random_step,
                }



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from abc import ABC, abstractclassmethod
import copy
import jsonlines
import random
import os
from typing import Any, Dict, List, Optional, Set, Tuple

from parlai.core.message import Message
from parlai.core.metrics import ExactMatchMetric, F1Metric
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import DialogTeacher
from parlai.core.torch_classifier_agent import ConfusionMatrixMetric
from parlai.utils.data import DatatypeHelper

from parlai.tasks.light_multiparty.build import build, DATASET_NAME

START_TOKEN = "__START__"
SILENCE_TOKEN = "__SILENCE__"
SPEAKER_TOKEN_DELIM = ":"


def _format_timestep(timestamp: str, first_utterance_timestamp: str):
    """
    Outputs timestep in the format of '00:00:00' relative to `first_utterance_timestamp`
    """
    time_delta = int(float(timestamp) - float(first_utterance_timestamp))
    hour = time_delta // 3600
    time_delta %= 3600
    minute = time_delta // 60
    time_delta %= 60
    second = time_delta
    return f"{hour:02d}:{minute:02d}:{second:02d}"


def get_clean_text(message: Message) -> str:
    return message["text"].replace("\n", " ")


def flatten_personas(
    personas: List[Dict[str, str]], delim: Optional[str] = "\n"
) -> str:
    personass_str_parts = []
    personass_str_parts.append("__personas__")
    personass_str_parts.extend([f"{p['name']}: {p['persona']}" for p in personas])
    personass_str_parts.append("__end-personas__")
    return delim.join(personass_str_parts)


def flatten_location(location: Dict[str, str], delim: Optional[str] = "\n") -> str:
    location_str_parts = [
        "__location__",
        f"{location['name']}: {location['description']}",
        "__end-location__",
    ]
    return delim.join(location_str_parts)


def get_speaker_names(utt: Message) -> List[str]:
    return [p["name"] for p in utt["personas"]]


def get_speaker_name_from_index(utt: Message, index: int) -> str:
    return utt["personas"][index]["name"]


class BaseTeacher(DialogTeacher):
    """
    Base class for MultiLIGHT teacher.

    This is an abstract class: do NOT use directly!
    """

    def __init__(self, opt: Opt, shared=None):
        opt = copy.deepcopy(opt)
        build(opt)
        # We often also call folds splits (for example in the build file here).
        self.fold = DatatypeHelper.fold(opt["datatype"])
        opt["datafile"] = os.path.join(
            opt["datapath"], DATASET_NAME, f"{self.fold}.jsonl"
        )

        self.episode_quality_tiers = self._get_data_quality_tiers(
            opt["episode_quality_tiers"]
        )
        self.speaker_quality_tiers = self._get_data_quality_tiers(
            opt["speaker_quality_tiers"]
        )
        self.utterance_delimiter = opt["utterance_delimiter"]
        self.use_start_token = opt["use_start_token"]
        self.start_token = opt["start_token"]
        self.include_speaker_in_context = opt["include_speaker_in_context"]
        self.include_speaker_in_label = opt["include_speaker_in_label"]
        self.add_speaker_to_context_end = opt["add_speaker_to_context_end"]
        self.speaker_token_delimiter = opt["speaker_token_delimiter"]
        self.include_timestep_in_context = opt["include_timestep_in_context"]
        self.add_current_timestep_to_context = opt["add_current_timestep_to_context"]
        self.add_personas_to_context = opt["add_personas_to_context"]
        self.add_location_to_context = opt["add_location_to_context"]
        self.id = "light_multiparty_dialogue"

        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group("MultiLIGHT Teacher Arguments")
        agent.add_argument(
            "--episode-quality-tiers",
            type=str,
            default="1,2",
            help="Comma separated list of tiers of data quality for episodes to include. "
            "The available tiers are 1 (high), 2 (low) quality."
            " NOTE: only training data split has tiers. valid and test are only tier 1 data.",
        )
        agent.add_argument(
            "--speaker-quality-tiers",
            type=str,
            default="1,2",
            help="Comma separated list of tiers of data quality for speakers to include. "
            "The available tiers are 1 (high), 2 (low) quality."
            " NOTE: only training data split has tiers. valid and test are only tier 1 data.",
        )
        agent.add_argument(
            "--utterance-delimiter",
            type=str,
            default="\n",
            help="A string used to separate each utterance in the context. Defaults to newline. For example, 'A: Hello\nB: Hi there'.",
        )
        agent.add_argument(
            "--use-start-token",
            type="bool",
            default=False,
            help="Use start token at the beginning of each conversation, and include the first sentence as a training example.",
        )
        agent.add_argument(
            "--start-token",
            type=str,
            default=START_TOKEN,
            help=f"The token to use to indicate the beginning of a conversation. Defaults to {START_TOKEN}",
        )
        agent.add_argument(
            "--include-speaker-in-label",
            type="bool",
            default=True,
            help="Whether to include speaker labels in the label. "
            "For example, message = { label: 'Rachel: Hi' } instead of message = { label: 'Hi' }",
        )
        agent.add_argument(
            "--include-speaker-in-context",
            type="bool",
            default=True,
            help="Whether to include speaker labels in the context. "
            "For example, message = { text: 'Rachel: Hi' } instead of message = { text: 'Hi' }",
        )
        agent.add_argument(
            "--add-speaker-to-context-end",
            type="bool",
            default=False,
            help="Append the current speaker (who says `label`) to the end of each context. Defaults to True.",
        )
        agent.add_argument(
            "--speaker-token-delimiter",
            type=str,
            default=SPEAKER_TOKEN_DELIM,
            help="The token to use to separate the speaker label from the actual utterance in `obs['text']`.",
        )
        agent.add_argument(
            "--include-timestep-in-context",
            type="bool",
            default=False,
            help="If true, will add the timestep as part of the context for each previous utterance.",
        )
        agent.add_argument(
            "--add-current-timestep-to-context",
            type="bool",
            default=False,
            help="If true, will add the current timestep at the end of the context.",
        )
        agent.add_argument(
            "--add-personas-to-context",
            type="bool",
            default=False,
            help="If true, will prepend the flattened persona descriptions to the context.",
        )
        agent.add_argument(
            "--add-location-to-context",
            type="bool",
            default=False,
            help="If true, will prepend the flattened location description to the context.",
        )
        return parser

    def _get_data_quality_tiers(self, tiers_str: str) -> Set[int]:
        tiers = set([int(t) for t in tiers_str.split(",")])
        for t in tiers:
            assert t in (
                1,
                2,
            ), f"Requested data tier ({t}) is invalid. Available tiers are 1 and 2 (only)."
        return tiers

    def get_speaker_prompt(self, utt: Message, ts: Optional[str] = None) -> str:
        spkr = utt["speaker"]
        return (
            f"{spkr} {ts} {self.speaker_token_delimiter}"
            if ts and self.include_timestep_in_context
            else f"{spkr}{self.speaker_token_delimiter}"
        )

    def get_utterance_context(self, utt: Message, ts: Optional[str] = None) -> str:
        text = get_clean_text(utt)
        if self.include_speaker_in_context:
            text = f"{self.get_speaker_prompt(utt, ts)} {text}"
        return text

    def get_utterance_label(self, utt: Dict[str, Any]) -> str:
        label = get_clean_text(utt)
        if self.include_speaker_in_label:
            label = f"{self.get_speaker_prompt(utt)} {label}"
        return label

    def setup_data(self, datafile: str) -> List[Message]:
        with jsonlines.open(datafile) as reader:
            conversations = [dl for dl in reader]

        assert conversations, "No data was loaded by teacher."

        for conv in conversations:
            if conv["quality_tier"] not in self.episode_quality_tiers:
                continue
            last_utterance_index = len(conv["messages"]) - 1

            first_utterance_timestamp = 0
            characters_index = dict()
            personas = []
            for i, c in enumerate(conv["characters"]):
                personas.append({"name": c["name"], "persona": c["persona"]})
                characters_index[c["name"]] = i + 1

            location = {
                "name": conv["location"]["name"],
                "description": conv["location"]["description"],
            }

            context = []

            for index, utterance in enumerate(conv["messages"]):
                speaker = utterance["speaker"]
                utterance["speaker_id"] = characters_index[speaker]
                if index == 0:
                    new_episode = True
                    # Setting the start time for the relative timestamps after this.
                    first_utterance_timestamp = utterance["timestamp"]

                    if self.use_start_token:
                        context.append(self.start_token)
                    else:
                        # If it is the first utterance,
                        # we just add it to the context for the next round.
                        context.append(
                            self.get_utterance_context(
                                utt=utterance, ts=_format_timestep(0, 0)
                            )
                        )
                        # skip the first utterance after creating the initial context.
                        continue

                isConversationDone = index == last_utterance_index

                timestamp = utterance["timestamp"]
                timestep = _format_timestep(timestamp, first_utterance_timestamp)

                yield {
                    "timestep": timestep,
                    "location": location,
                    "personas": personas,
                    "speaker": speaker,
                    "characters": [p["name"] for p in personas],
                    "speaker_id": utterance["speaker_id"],
                    "full_context": context,
                    "speaker_worker_tier": conv['workers_quality_check'][
                        utterance["speaker_id"] - 1
                    ]["worker_tier"],
                    "quality_tier": conv["quality_tier"],
                    "text": None,  # This is an abstract class, we will feel in the text later.
                    "label": self.get_utterance_label(utterance),
                }, new_episode

                new_episode = False
                # Generating context for the next round.
                if not isConversationDone:
                    context.append(
                        self.get_utterance_context(utt=utterance, ts=timestep)
                    )

    def get_utterance_speaker_id(self, personas: Dict[str, str], speaker: str):
        for idx, persona in enumerate(personas):
            if persona["name"] == speaker:
                return idx

    def get_extra_context_before(self, conv: Message):
        """
        Generates the persona and location, which goes *before* the conversation
        context.
        """
        extra_context_before = []

        if self.add_location_to_context:
            extra_context_before.append(flatten_location(conv["location"]))
        if self.add_personas_to_context:
            extra_context_before.append(flatten_personas(conv["personas"]))
        return extra_context_before

    def get_extra_context_after(self, conv: Message):
        """
        Generates the timestep and speaker prompt, which goes *after* the conversation
        context.
        """
        extra_context_after = []
        if self.add_current_timestep_to_context:
            extra_context_after.append(conv["timestep"])
        if self.add_speaker_to_context_end:
            # Adding the speaker prompt to the end of the context
            extra_context_after.append(self.get_speaker_prompt(conv))
        return extra_context_after


class AllSpeakersTeacher(BaseTeacher):
    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.id = "multilight_dialogue_:all_speakers"

    def setup_data(self, datafile: str) -> List[Message]:
        for utt, _ in super().setup_data(datafile):
            ret = copy.deepcopy(utt)
            extra_context_before = self.get_extra_context_before(ret)
            extra_context_after = self.get_extra_context_after(ret)

            ret["text"] = self.utterance_delimiter.join(
                extra_context_before + ret["full_context"] + extra_context_after
            )
            if utt["speaker_worker_tier"] in self.speaker_quality_tiers:
                yield ret, True

    def custom_evaluation(
        self,
        teacher_action: Message,
        labels: Optional[Tuple[str]],
        model_response: Message,
    ) -> None:
        model_response = Message(model_response)
        if model_response.is_padding() or (not model_response.get("text", None)):
            return

        # Finding speaker label accuracy, only if it was included in the teacher labels
        if self.include_speaker_in_label:
            resp = model_response.get("text")
            parts = resp.split(self.speaker_token_delimiter)

            # If model predicts a speaker label, there should be at leat two parts separted by ':'
            if len(parts) >= 2:
                predited_speaker = parts[0].strip()
            else:
                # A dummy string to trigger this as the wrong speaker.
                predited_speaker = "__NO_SPEAKER__"

            # The teacher will always include the correct speaker label in the 'speaker' field
            expected_speaker = teacher_action.get("speaker")
            # Predicted speaker accuracy
            self.metrics.add(
                "speaker_acc",
                ExactMatchMetric.compute(predited_speaker, [expected_speaker]),
            )

            # unigram F1 Metric for the speech part only, removing the speaker token
            speech_text = parts[-1]
            label_speech_text = labels[0].split(self.speaker_token_delimiter)[-1]
            self.metrics.add(
                "speech_f1",
                F1Metric.compute(speech_text, [label_speech_text]),
            )


class SpeakerPredictionTeacher(AllSpeakersTeacher):
    """
    Generates the speaker name for the current turn.
    """

    def __init__(self, opt: Opt, shared=None):
        self.add_current_turn = opt["add_current_turn"]
        super().__init__(opt, shared)
        self.id = "SpeakerPrediction"

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group("MultiLIGHT Speaker prediction.")
        agent.add_argument(
            "--add-current-turn",
            type="bool",
            default=False,
            help="Whether to include the current turn in the text.",
        )

    def _label_text(self, message: Message):
        """
        Removes the speaker when the last utterance is being added to the context.
        """
        label = message["label"]
        if self.include_speaker_in_label:
            # Just a quick check to be extra sure about the data format.
            label_speaker, label_text = label.split(self.speaker_token_delimiter, 1)
            assert (
                label_speaker.strip() == message["speaker"]
            ), "Wrong speaker in the label."
            label = label_text.strip()
        return label

    def setup_data(self, datafile: str) -> List[Message]:
        for utt, _e in super().setup_data(datafile):
            msg = copy.deepcopy(utt)
            current_turn_utterance = self._label_text(msg)
            msg["label"] = msg["speaker"]
            msg["label_candidates"] = get_speaker_names(msg)
            if self.add_current_turn:
                msg["text"] = self.utterance_delimiter.join(
                    [msg["text"], current_turn_utterance]
                )
            yield msg, _e


class SingleSpeakerTeacher(ABC, BaseTeacher):
    """
    Generaes the utterances for a single character only.

    The label will be the silent token if that character doesn't speak that round. This
    is an abstract class: do NOT use directly!
    """

    def __init__(self, opt: Opt, shared=None):
        self.silence_token = opt["silence_token"]
        self.silence_token_dropout = opt["silence_token_dropout"]
        assert (
            0 <= self.silence_token_dropout <= 1
        ), "--silence-token-dropout must be a number in [0, 1]"
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group(
            "MultiLIGHT Teacher Arguments (Single Speaker Teacher)"
        )
        agent.add_argument(
            "--silence-token",
            type=str,
            default=SILENCE_TOKEN,
            help=f"The token to use to indicate the chosen speaker is silent. Defaults to {SILENCE_TOKEN}",
        )
        agent.add_argument(
            "--silence-token-dropout",
            type=float,
            default=0,
            help="Dropout probability for using silence token to generate training example"
            " for sentences where the chosen speaker is not speaking. "
            "When set to 0, all silence tokens will generate training examples. "
            "When set to 1, no silence tokens will generate training examples.",
        )
        return parser

    @abstractclassmethod
    def get_speaker_index(self) -> int:
        """
        Returns the index of the speaker in the dataset (0, 1, 2)
        """

    def generate_silence_label(self, speaker: str) -> str:
        label = self.silence_token
        if self.include_speaker_in_label:
            label = f"{speaker}{self.speaker_token_delimiter} {label}"
        return label

    def is_this_speaker_turn(self, utt: Message) -> bool:
        return (
            self.get_utterance_speaker_id(utt["personas"], utt["speaker"])
            == self.get_speaker_index()
        )

    def setup_data(self, datafile: str) -> List[Message]:
        episode_needs_reset = False

        for utt, new_episode in super().setup_data(datafile):
            # checking if we need to skip this round because of the speaker tier
            if utt["speaker_worker_tier"] not in self.speaker_quality_tiers:
                continue

            if new_episode:
                last_speaker_trun = -1
                turn_num = 0
            else:
                turn_num += 1

            is_silence_turn = not self.is_this_speaker_turn(utt)
            if (
                is_silence_turn
                and self.silence_token_dropout
                and random.random() < self.silence_token_dropout
            ):
                if new_episode:
                    # We keep this to reset the episode next round around.
                    episode_needs_reset = True
                continue

            if episode_needs_reset:
                # Setting new episode because we missed the original one due to silence drop out.
                episode_needs_reset = False
                new_episode = True

            ret = copy.deepcopy(utt)
            context = ret["full_context"][(last_speaker_trun + 1) :]

            if not context:
                context = [self.silence_token] if turn_num > 0 else [self.start_token]

            if last_speaker_trun < 0:
                # Only adding the extra context on the first episode that will get yielded
                ret["text"] = self.utterance_delimiter.join(
                    self.get_extra_context_before(ret) + context
                )
            else:
                ret["text"] = self.utterance_delimiter.join(context)

            last_speaker_trun = turn_num
            # Setting the speaker to the current speaker.
            # Even if the speaker doesn't speak, the silence term is this speaker's action.
            ret["speaker"] = get_speaker_name_from_index(utt, self.get_speaker_index())

            # Adding the prompot to the temp_history for not keeping it in the full history.
            # There is also an extra \n to make the format more alike the other teachers.
            ret["temp_history"] = "\n".join([""] + self.get_extra_context_after(ret))
            if is_silence_turn:
                ret["label"] = self.generate_silence_label(ret["speaker"])
            else:
                last_speaker_trun += 1

            yield ret, new_episode

    def custom_evaluation(
        self,
        teacher_action: Message,
        labels: Optional[Tuple[str]],
        model_response: Message,
    ) -> None:
        if model_response.is_padding() or (not model_response.get("text", None)):
            return

        model_predicted_speak = str(self.silence_token not in model_response["text"])
        label_was_speak = str(self.silence_token not in labels[0])
        self.metrics.add(
            "self_speaker_acc",
            ExactMatchMetric.compute(model_predicted_speak, [label_was_speak]),
        )
        # Precited speaker confusion metrics
        precision, recall, f1 = ConfusionMatrixMetric.compute_metrics(
            [model_predicted_speak], [label_was_speak], "True"
        )
        self.metrics.add("self_speaker_precision", precision[0])
        self.metrics.add("self_speaker_recall", recall[0])
        self.metrics.add("self_speaker_f1", f1[0])


class FirstSpeakerTeacher(SingleSpeakerTeacher):
    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.id = "multilight_dialogue:first_speaker"

    def get_speaker_index(self) -> int:
        return 0


class SecondSpeakerTeacher(SingleSpeakerTeacher):
    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.id = "multilight_dialogue:second_speaker"

    def get_speaker_index(self) -> int:
        return 1


class ThirdSpeakerTeacher(SingleSpeakerTeacher):
    def __init__(self, opt: Opt, shared=None):
        super().__init__(opt, shared)
        self.id = "multilight_dialogue:third_speaker"

    def get_speaker_index(self) -> int:
        return 2


class DefaultTeacher(AllSpeakersTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Example Flask server which hosts a model.

## Examples
**Serving the model**
```shell
parlai flask -m repeat_query
parlai flask -mf zoo:blender/blender_90M/model
```

**Hitting the API***
```shell
curl -k http://localhost:5000/response -H "Content-Type: application/json" -d '{"text": "foobar"}'
```
"""

from parlai.core.agents import create_agent
from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script


@register_script('flask', hidden=True)
class Flask(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = ParlaiParser(True, True)
        return parser

    def chatbot_response(self):
        from flask import request

        data = request.json
        self.agent.observe({'text': data["text"], 'episode_done': False})
        response = self.agent.act()
        return {'response': response['text']}

    def run(self):
        from flask import Flask

        self.agent = create_agent(self.opt)
        app = Flask("parlai_flask")
        app.route("/response", methods=("GET", "POST"))(self.chatbot_response)
        app.run()


if __name__ == "__main__":
    Flask.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
The speaker prediction model for MultiLIGHT.
"""


from parlai.zoo.multilight.utils import maybe_download

MODEL_NAME = 'speaker'


def download(datapath):
    maybe_download(MODEL_NAME, datapath)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Mutators for the SeeKeR Tasks.
"""
from abc import abstractmethod
import nltk
from parlai.core.message import Message
from parlai.core.mutators import register_mutator, MessageMutator, ManyEpisodeMutator
from parlai.tasks.wizard_of_internet import constants as CONST
import parlai.tasks.wizard_of_internet.mutators  # type: ignore
import parlai.utils.logging as logging

try:
    import spacy
except ModuleNotFoundError:
    logging.error('Please install spacy: pip install spacy')
    spacy = None

from projects.seeker.utils import (
    GENERATE_QUERY,
    IS_SEARCH_REQUIRED,
    DO_SEARCH,
    DO_NOT_SEARCH,
    extract_entities,
    calc_f1_msmarco,
    calc_f1_msc,
)
from parlai.tasks.wizard_of_wikipedia.agents import TOKEN_KNOWLEDGE, TOKEN_END_KNOWLEDGE

nlp = None

IGNORE_ENTITY_TYPES = set(
    ['CARDINAL', 'DATE', 'ORDINAL', 'PERCENT', 'QUANTITY', 'TIME']
)


@register_mutator('skip_retrieval_mutator')
class SkipRetrievalMutator(MessageMutator):
    """
    Mutator that adds a 'skip_retrieval' key to the observation.
    """

    def message_mutation(self, message: Message) -> Message:
        message.force_set('skip_retrieval', True)
        return message


@register_mutator('add_selected_sentences_mutator')
class AddSelectedSentencesMutator(MessageMutator):
    """
    Mutator that adds selected sentences to the messages.
    """

    def message_mutation(self, message: Message) -> Message:
        if 'checked_sentence' in message:
            message[CONST.SELECTED_SENTENCES] = [message['checked_sentence']]
        elif CONST.SELECTED_SENTENCES not in message:
            message[CONST.SELECTED_SENTENCES] = []
        return message


@register_mutator('add_retrieved_documents_mutator')
class AddRetrievedDocumentsMutator(MessageMutator):
    """
    Add retrieved docs and relevant keys.
    """

    def message_mutation(self, message: Message) -> Message:
        sentences = message['text'].split('\n')[:-1]
        if CONST.RETRIEVED_DOCS not in message:
            message[CONST.RETRIEVED_DOCS] = sentences
        if CONST.SELECTED_SENTENCES not in message:
            message[CONST.SELECTED_SENTENCES] = message['labels']
        if CONST.RETRIEVED_DOCS_URLS not in message:
            message[CONST.RETRIEVED_DOCS_URLS] = [''] * len(sentences)
        if CONST.RETRIEVED_DOCS_TITLES not in message:
            message[CONST.RETRIEVED_DOCS_TITLES] = [''] * len(sentences)
        return message


@register_mutator('prompt_search_query_mutator')
class PromptSearchQueryMutator(MessageMutator):
    """
    Add a __generate_search_query__ prompt to the end of the context, to inform the
    model.

    Assumes flattened data.
    """

    PROMPT = GENERATE_QUERY

    def message_mutation(self, message: Message) -> Message:
        if not message['text'].endswith(self.PROMPT):
            message.force_set('text', f"{message['text']} {self.PROMPT}")
        return message


@register_mutator('strip_context_mutator')
class StripContextMutator(MessageMutator):
    """
    Strip the context.

    This can be used to turn QA tasks to "open" QA tasks.
    """

    def message_mutation(self, message: Message) -> Message:
        message.force_set('text', message['text'].split('\n')[-1])
        return message


class SearchQueryClassificationMixin(MessageMutator):
    """
    Changes the message in the following ways:

    1. Makes the task *only one line of context*
    2. Adds a prompt to the end of the message, indicating a binary choice of __search-required__
    3. Changes the label to indicate whether to search or not.
    """

    PROMPT = IS_SEARCH_REQUIRED
    LABEL: str

    def message_mutation(self, message: Message) -> Message:
        assert self.get_label(message)
        if not message['text'].endswith(self.PROMPT):
            last_context = message['text'].split('\n')[-1]
            message.force_set('text', f"{last_context} {self.PROMPT}")
        if message['labels'] != [self.get_label(message)]:
            message.force_set('labels', [self.get_label(message)])
        message.pop('knowledge', None)
        return message

    @abstractmethod
    def get_label(self, message: Message) -> str:
        """
        Return the label.
        """


@register_mutator('do_generate_search_query_mutator')
class DoGenerateSearchQueryMutator(SearchQueryClassificationMixin):
    def get_label(self, message: Message) -> str:
        return DO_SEARCH


@register_mutator('dont_generate_search_query_mutator')
class DontGenerateSearchQueryMutator(SearchQueryClassificationMixin):
    def get_label(self, message: Message) -> str:
        return DO_NOT_SEARCH


@register_mutator('wow_maybe_generate_search_query_mutator')
class WowMaybeGenerateSearchQueryMutator(SearchQueryClassificationMixin):
    def get_label(self, message: Message) -> str:
        checked_sentence = message.get('checked_sentence')
        return DO_NOT_SEARCH if checked_sentence == 'no_passages_used' else DO_SEARCH


@register_mutator('woi_maybe_generate_search_query_mutator')
class WoiMaybeGenerateSearchQueryMutator(SearchQueryClassificationMixin):
    def get_label(self, message: Message) -> str:
        checked_sentences = ' '.join(message.get(CONST.SELECTED_SENTENCES))
        return (
            DO_NOT_SEARCH
            if checked_sentences == CONST.NO_SELECTED_SENTENCES_TOKEN
            else DO_SEARCH
        )


@register_mutator('bst_tasks_maybe_generate_search_query_mutator')
class BSTMaybeGenerateSearchQueryMutator(SearchQueryClassificationMixin):
    def get_label(self, message: Message) -> str:
        global nlp
        if nlp is None:
            assert spacy is not None
            nlp = spacy.load('en_core_web_sm')
        entities = nlp(message['text'].split('\n')[-1]).ents
        if entities:
            entities = [e for e in entities if e.label_ not in IGNORE_ENTITY_TYPES]
        return DO_NOT_SEARCH if not entities else DO_SEARCH


@register_mutator("woi_pop_documents_mutator")
class WoiRemoveDocuments(MessageMutator):
    """
    Remove documents from all message fields.

    Reduces space taken by the data.
    """

    def message_mutation(self, message: Message) -> Message:
        message.force_set(CONST.RETRIEVED_DOCS, [''])
        message.force_set(CONST.RETRIEVED_SENTENCES, [''])
        message.force_set(CONST.RETRIEVED_DOCS_TITLES, [''])
        message.force_set(CONST.RETRIEVED_DOCS_URLS, [''])
        message.force_set(CONST.SELECTED_DOCS, [''])
        message.force_set(CONST.SELECTED_DOCS_TITLES, [''])
        message.force_set(CONST.SELECTED_DOCS_URLS, [''])
        return message


@register_mutator("ms_marco_to_woi")
class MsMarcoToWoi(MessageMutator):
    """
    Turns context into docs amenable to FiD training.
    """

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        if not isinstance(new_message[CONST.RETRIEVED_DOCS], list):
            new_message.force_set(
                CONST.RETRIEVED_DOCS, [new_message[CONST.RETRIEVED_DOCS]]
            )
        new_message[CONST.RETRIEVED_DOCS_TITLES] = [''] * len(
            new_message[CONST.RETRIEVED_DOCS]
        )
        new_message[CONST.RETRIEVED_DOCS_URLS] = [''] * len(
            new_message[CONST.RETRIEVED_DOCS]
        )
        return new_message


@register_mutator("ms_marco_filter_has_answer")
class MsMarcoFilterHasAnswer(ManyEpisodeMutator):
    """
    Filters out examples that do not have an answer present.
    """

    def many_episode_mutation(self, episode):
        out_episodes = []
        for e in episode:
            if e.get('labels')[0] != 'No Answer Present.':
                out_episodes.append([e])
        return out_episodes


@register_mutator("ms_marco_create_fid_docs")
class MsMarcoCreateFidDocs(MessageMutator):
    """
    Turns context into docs amenable to FiD training.
    """

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        docs = message['text'].split('\n')
        context = docs[-1]
        docs = docs[0:-1]
        new_message[CONST.RETRIEVED_DOCS] = docs
        new_message.force_set('text', context)
        return new_message


@register_mutator("ms_marco_find_selected_sentence_for_knowledge")
class MsMarcoFindSelectedSentenceKnowledge(ManyEpisodeMutator):
    """
    Finds selected sentence for use in K2R.
    """

    target = 'knowledge'

    def many_episode_mutation(self, episode):
        out_episodes = []
        for m in episode:
            new_m = m.copy()
            gold_label = m.get('labels', [''])[0]
            try:
                gold_label_parts = nltk.word_tokenize(gold_label)
            except IndexError:
                # malformed gold label; continue
                continue
            best_f1 = 0
            best_sentence = ''
            docs = m.get(CONST.RETRIEVED_DOCS)
            for d in docs:
                try:
                    ds = nltk.sent_tokenize(d)
                except IndexError:
                    # malformed documents; continue as if it does not exist.
                    continue
                for s in ds:
                    f1 = calc_f1_msmarco(s, gold_label_parts)
                    if f1 > best_f1:
                        best_f1 = f1
                        best_sentence = s
            ms_marco_f1_threshold = 0.5
            if self.target == 'knowledge':
                new_m.force_set('labels', [best_sentence])
            else:
                new_m.force_set(
                    'text',
                    new_m['text']
                    + f'\n{TOKEN_KNOWLEDGE} '
                    + best_sentence
                    + f' {TOKEN_END_KNOWLEDGE}',
                )

            new_m.force_set(CONST.SELECTED_SENTENCES, [best_sentence])
            if best_f1 > ms_marco_f1_threshold:
                out_episodes.append([new_m])
        return out_episodes


@register_mutator("ms_marco_find_selected_sentence_for_response")
class MsMarcoFindSelectedSentenceResponse(MsMarcoFindSelectedSentenceKnowledge):
    target = 'response'


@register_mutator("squad_to_woi")
class SquadToWoi(MessageMutator):
    """
    Turns context into docs amenable to FiD training.
    """

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        docs = message['text'].split('\n')
        context = docs[-1]
        docs = docs[0:-1]
        new_message[CONST.RETRIEVED_DOCS] = docs
        new_message[CONST.RETRIEVED_DOCS_TITLES] = [''] * len(docs)
        new_message[CONST.RETRIEVED_DOCS_URLS] = [''] * len(docs)
        new_message.force_set('text', context)
        return new_message


@register_mutator("triviaqa_to_woi")
class TriviaqaToWoi(MessageMutator):
    """
    Turns context into docs amenable to FiD training.
    """

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        docs = message['text'].split('\n')
        context = docs[-1]
        docs = [' '.join(docs[0:-1])]
        new_message[CONST.RETRIEVED_DOCS] = docs
        new_message[CONST.RETRIEVED_DOCS_TITLES] = ['']
        new_message[CONST.RETRIEVED_DOCS_URLS] = ['']
        new_message.force_set('text', context)
        return new_message


@register_mutator("nqopen_to_woi")
class NQOpenToWoi(MessageMutator):
    """
    Turns context into docs amenable to FiD training.
    """

    def message_mutation(self, message: Message) -> Message:
        new_message = message.copy()
        new_message[CONST.RETRIEVED_DOCS] = [message['checked_sentence']]
        new_message[CONST.RETRIEVED_DOCS_TITLES] = [message['title']]
        new_message[CONST.RETRIEVED_DOCS_URLS] = ['']
        new_message.pop('checked_sentence')
        if message.get('history', '') != '':
            text = message.get('history', '') + '\n' + message.get('text', '')
            new_message.force_set('text', text)
        return new_message


@register_mutator("nq_to_woi")
class NQToWoi(ManyEpisodeMutator):
    """
    Turns context into docs amenable to FiD training.

    Note we currently throw away all examples with no label.
    """

    def many_episode_mutation(self, episode):
        out_episodes = []
        for message in episode:
            if message['labels'][0] == '':
                pass
            else:
                new_message = message.copy()
                docs = message['text'].split('\n')
                context = docs[-1]
                docs = docs[0:-1]
                new_message[CONST.RETRIEVED_DOCS] = docs
                new_message[CONST.RETRIEVED_DOCS_TITLES] = ['']
                new_message[CONST.RETRIEVED_DOCS_URLS] = ['']
                new_message.force_set('text', context)
                out_episodes.append([new_message])
        return out_episodes


@register_mutator("extract_entity_for_knowledge_model")
class ExtractEntity(ManyEpisodeMutator):
    """
    Picks an entity in the context and uses it as the intended target.
    """

    def many_episode_mutation(self, episode):
        out_episodes = []
        for message in episode:
            new_message = message.copy()
            context_except_last_line = '\n'.join(message['text'].split('\n')[:-1])
            context_ent = extract_entities(context_except_last_line)
            label_ent = extract_entities(message['labels'][0])
            ents = set(context_ent).intersection(label_ent)
            if len(ents) > 0:
                longest_ent = max(ents, key=len)
            else:
                continue
            new_message.force_set('response_labels', message['labels'])
            new_message.force_set('labels', [longest_ent])
            new_message.force_set(CONST.SELECTED_SENTENCES, [longest_ent])
            new_message.force_set(
                CONST.RETRIEVED_DOCS, message['text'].split('\n')[:-1]
            )
            blanks = [''] * len(message['text'].split('\n')[:-1])
            new_message.force_set(CONST.RETRIEVED_DOCS_URLS, blanks)
            new_message.force_set(CONST.RETRIEVED_DOCS_TITLES, blanks)
            out_episodes.append([new_message])
        return out_episodes


@register_mutator("extract_entity_for_response_model")
class ExtractEntityResponse(ManyEpisodeMutator):
    """
    Picks an entity in the context and uses it as the intended target.
    """

    BEGIN_KNOWLEDGE: str = CONST.KNOWLEDGE_TOKEN
    END_KNOWLEDGE: str = CONST.END_KNOWLEDGE_TOKEN

    def many_episode_mutation(self, episode):
        out_episodes = []
        for message in episode:
            new_message = message.copy()
            context_except_last_line = '\n'.join(message['text'].split('\n')[:-1])
            context_ent = extract_entities(context_except_last_line)
            label_ent = extract_entities(message['labels'][0])
            ents = set(context_ent).intersection(label_ent)
            if len(ents) > 0:
                longest_ent = max(ents, key=len)
            else:
                continue
            knol_text = f'{self.BEGIN_KNOWLEDGE} {longest_ent} {self.END_KNOWLEDGE}'
            new_text = message['text'] + '\n' + knol_text
            new_message.force_set('text', new_text)
            out_episodes.append([new_message])
        return out_episodes


@register_mutator("msc_find_selected_sentence_knowledge")
class MscFindSelectedSentenceKnowledge(ManyEpisodeMutator):
    """
    Finds selected sentence for use in K2R.
    """

    target = 'knowledge'

    def many_episode_mutation(self, episode):
        out_episodes = []
        for m in episode:
            new_m = m.copy()
            gold_label = m.get('labels', [''])[0]
            try:
                gold_label_parts = nltk.word_tokenize(gold_label)
            except IndexError:
                # malformed label
                gold_label_parts = []
            best_f1 = 0
            best_sentence = ''
            context_except_last_line = '\n'.join(m['text'].split('\n')[:-1])
            docs = [context_except_last_line]
            for d in docs:
                ds = d.split('\n')
                for s in ds:
                    f1 = calc_f1_msc(s, gold_label_parts)
                    if f1 > best_f1:
                        best_f1 = f1
                        best_sentence = s
            if best_sentence != '':
                # find which speaker it is
                z = m['text'].split('\n')
                z.reverse()
                ind = z.index(best_sentence)
                if (ind % 2) == 0:
                    speaker = '__them__'
                else:
                    speaker = '__you__'
                if 'your persona:' in best_sentence:
                    speaker = '__you__'
                best_sentence = best_sentence + " " + speaker
            msc_f1_threshold = 0.3
            if self.target == 'knowledge':
                new_m.force_set('labels', [best_sentence])
            else:
                new_m.force_set(
                    'text',
                    new_m['text']
                    + '\n__knowledge__ '
                    + best_sentence
                    + ' __endknowledge__',
                )
            new_m.force_set('old_target', gold_label)

            if best_f1 > msc_f1_threshold:
                out_episodes.append([new_m])
        return out_episodes


@register_mutator("msc_find_selected_sentence_response")
class MscFindSelectedSentenceResponse(MscFindSelectedSentenceKnowledge):
    """
    Finds selected sentence for use in K2R.
    """

    target = 'response'



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Multiwoz 2.2 Dataset implementation for ParlAI.
"""

import copy
import json
import os
from typing import Optional

import numpy as np
import pandas as pd

import parlai.core.tod.tod_agents as tod_agents
import parlai.core.tod.tod_core as tod
import parlai.tasks.multiwoz_v22.build as build_
from parlai.core.message import Message
from parlai.core.metrics import AverageMetric
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.utils.data import DatatypeHelper
from parlai.utils.io import PathManager
from parlai.utils.testing import is_this_circleci

DOMAINS = [
    "attraction",
    "bus",
    "hospital",
    "hotel",
    "police",
    "restaurant",
    "taxi",
    "train",
]

WELL_FORMATTED_DOMAINS = ["attraction", "bus", "hotel", "restaurant", "train", "taxi"]

DATA_LEN = {"train": 17, "dev": 2, "test": 2}

SEED = 42


def fold_size(fold):
    return DATA_LEN[fold]


class MultiwozV22Parser(tod_agents.TodStructuredDataParser):
    """
    Abstract data loader for Multiwoz V2.2 into TOD structured data format.

    Multiwoz 2.2 has 'find' and 'book' as the only intents.

    For API calls, we look for turns that are not 'NONE' `active_intents` in the USER's turn state. We then filter these for whether or not the SYSTSEM has actually made an api call by looking in the dialogue act of the SYSTEM turn.
    * For 'find' intents, we make an API call if it does an "Inform" or gives a "NoOffer". We look in the corresponding `.db` file to return the relevant information.
    * For 'book' intents, we make an API call if the SYSTEM's dialogue act includes  booking and then offer the slots/values of that key as the API response.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        group = parser.add_argument_group("Multiwoz args")
        group.add_argument(
            "--well-formatted-domains-only",
            type=bool,
            default=True,
            help="Some of the domains in Multiwoz are not super well formatted. Use only the well formatted ones.",
        )
        group.add_argument(
            "--dialogue-id",
            type=str,
            default="",
            help="If non-empty, filters for a particular dialogue id",
        )
        return super().add_cmdline_args(parser, partial_opt)

    def __init__(self, opt: Opt, shared=None):
        self.fold = DatatypeHelper.fold(opt["datatype"])
        opt["datafile"] = self.fold
        self.dpath = os.path.join(opt["datapath"], "multiwoz_v22")
        build_.build(opt)
        self.last_call = {}
        super().__init__(opt, shared)

    def load_schemas(self):
        with PathManager.open(os.path.join(self.dpath, "schema.json")) as f:
            raw = json.load(f)
        result = {}
        for service in raw:
            domain = service["service_name"]
            prefix_end_idx = len(domain) + 1
            all_slots = set([x["name"][prefix_end_idx:] for x in service["slots"]])

            for intent in service["intents"]:
                call_name = intent["name"]
                result[call_name] = {tod.STANDARD_API_NAME_SLOT: call_name}
                req_slots = set([x[prefix_end_idx:] for x in intent["required_slots"]])
                if len(req_slots) > 0:
                    result[call_name][tod.STANDARD_REQUIRED_KEY] = list(req_slots)

                # Not fully trusting the original schema data...
                optional_slots = set(
                    [x[prefix_end_idx:] for x in intent["optional_slots"].keys()]
                )
                optional_slots = optional_slots | all_slots
                optional_slots = optional_slots - req_slots
                if len(optional_slots) > 0:
                    result[call_name][tod.STANDARD_OPTIONAL_KEY] = list(optional_slots)

            if domain == "police":  # Multiwoz 2.2 only lists "police"
                result["find_police"] = {
                    tod.STANDARD_OPTIONAL_KEY: list(all_slots),
                    tod.STANDARD_API_NAME_SLOT: "find_police",
                }
            if (
                domain == "taxi"
            ):  # Multiwoz 2.2 has "book taxi" in the schema but it's "find taxi" in the data...
                result["find_taxi"] = copy.deepcopy(result["book_taxi"])
                result["find_taxi"][tod.STANDARD_API_NAME_SLOT] = "find_taxi"
        return result

    def load_dbs(self):
        dbs = {}
        for key in DOMAINS:
            if (
                key == "hospital"
            ):  # has funky extra format, so we're gonna deal with it manually.
                with PathManager.open(
                    os.path.join(self.dpath, "db", key + "_db.json")
                ) as f:
                    file_lines = f.readlines()
                hospital_address_lines = file_lines[1:4]
                partial = [
                    x.replace("#", "").strip().lower().split(":")
                    for x in hospital_address_lines
                ]
                self.hospital_address = {x[0]: x[1] for x in partial}
                self.hospital_department_details = json.loads("".join(file_lines[6:]))
                continue
            if (
                key == "taxi"
            ):  # Taxi domain is funky and the db for it is just response slot options.
                continue
            with PathManager.open(
                os.path.join(self.dpath, "db", key + "_db.json")
            ) as f:
                blob = json.load(f)
                for i, entry in enumerate(blob):
                    cased = {}
                    for slot_name in entry:
                        cased[slot_name.lower().replace(" ", "")] = entry[slot_name]
                    blob[i] = cased
                dbs[key] = pd.DataFrame.from_dict(blob)

        return dbs

    def load_chunks(self, fold):
        if fold == "valid":
            fold = "dev"  # change name to match file structure
        for path in PathManager.ls(os.path.join(self.dpath, fold)):
            with PathManager.open(os.path.join(self.dpath, fold, path)) as f:
                blob = json.load(f)
                for convo in blob:
                    yield convo

    def _get_find_api_response(self, intent, raw_slots, sys_dialog_act):
        """
        Get an API response out of the lookup databases.
        """
        domain = ""
        for cand in DOMAINS:
            if cand in intent:
                domain = cand
        if domain == "taxi":  # handle separately cause funky
            for action in sys_dialog_act:
                if action == "Taxi-Inform":
                    return {x[0]: x[1] for x in sys_dialog_act[action]}
            return {domain: domain}  # too much work to do this right...
        if domain == "hospital":  # handle separately cause funky
            res = self.hospital_address
            if "hospital-department" in raw_slots:
                for blob in self.hospital_department_details:
                    if blob["department"] in raw_slots["hospital-department"]:
                        res[blob["department"]] = blob
            return res
        slots = {}
        for raw_key in raw_slots:
            key = raw_key[len(domain + "-") :]
            slots[key] = raw_slots[raw_key]
        for action in sys_dialog_act:
            if "Recommend" in action:
                add_slots = {}
                for x in sys_dialog_act[action]:
                    name = x[0]
                    val = x[1]
                    if self._slot_in_schema(name, intent):
                        if name not in add_slots:
                            add_slots[name] = []
                        add_slots[name].append(val)
                for key in add_slots:
                    slots[key] = add_slots[key]

        find = self.dbs[domain]
        for slot, values in slots.items():
            if slot == "arriveby":
                condition = find[slot] < values[0]
            elif slot == "leaveat":
                condition = find[slot] > values[0]
            else:
                condition = find[slot].isin(values)

            find = find[condition]

        filtered = self.dbs[domain].iloc[find.index]
        count = len(filtered.index)
        if count == 0:
            return {}
        blob = filtered.head(1).to_dict('records')

        results = blob[0]
        results["COUNT"] = count
        return results

    def _slot_in_schema(self, slot, intent):
        return slot in self.schemas[intent].get(
            tod.STANDARD_OPTIONAL_KEY, []
        ) or slot in self.schemas[intent].get(tod.STANDARD_REQUIRED_KEY, [])

    def _get_round(self, dialogue_id, raw_episode, turn_id):
        """
        Parse to TodStructuredRound.

        Assume User turn first.
        """
        user_turn = raw_episode[turn_id]
        if user_turn["speaker"] != "USER":
            raise RuntimeError(
                f"Got non-user turn when it should have been in {dialogue_id}; turn id {turn_id}"
            )
        sys_turn = raw_episode[turn_id + 1]
        sys_dialog_act = self.dialog_acts[dialogue_id][str(turn_id + 1)]["dialog_act"]
        if sys_turn["speaker"] != "SYSTEM":
            raise RuntimeError(
                f"Got non-system turn when it should have been in {dialogue_id}; turn id {turn_id}"
            )
        frames = user_turn.get("frames", [])
        call = {}
        resp = {}
        for frame in frames:
            if frame.get("state", {}).get("active_intent", "NONE") != "NONE":
                intent = frame["state"]["active_intent"]
                domain = frame["service"]
                maybe_call_raw = copy.deepcopy(frame["state"]["slot_values"])
                maybe_call = {}
                truncate_length = len(domain) + 1
                for key in maybe_call_raw:
                    maybe_call[key[truncate_length:]] = maybe_call_raw[key][0]
                maybe_call[tod.STANDARD_API_NAME_SLOT] = intent
                if "find" in intent:
                    for key in sys_dialog_act:
                        if "Inform" in key or "NoOffer" in key:
                            # Gotta check to make sure if it's inform, that it's about the right topic
                            if "Inform" in key:
                                valid = True
                                slots = [x[0] for x in sys_dialog_act[key]]
                                for slot in slots:
                                    valid &= self._slot_in_schema(slot, intent) | (
                                        slot == "choice"
                                    )
                                if not valid:
                                    continue

                            call = maybe_call
                            call_key = str(call)
                            if call_key not in self.call_response_cache:
                                resp = self._get_find_api_response(
                                    intent,
                                    frame["state"]["slot_values"],
                                    sys_dialog_act,
                                )
                                self.call_response_cache[call_key] = resp
                            else:
                                resp = self.call_response_cache[call_key]

                elif "book" in intent:
                    for key in sys_dialog_act:
                        if "Book" in key:  # and "Inform" not in key:
                            resp = {x[0]: x[1] for x in sys_dialog_act[key]}
                            call = maybe_call
        if call == self.last_call:
            call = {}
            resp = {}
        if len(call) > 0:
            self.last_call = call
        return (
            call,
            tod.TodStructuredRound(
                user_utt=user_turn["utterance"],
                api_call_machine=call,
                api_resp_machine=resp,
                sys_utt=sys_turn["utterance"],
            ),
        )

    def _get_schemas_for_goal_calls(self, goals):
        result = []
        seen = set()
        for goal in goals:
            call_name = goal[tod.STANDARD_API_NAME_SLOT]
            if call_name not in seen:
                result.append(self.schemas[call_name])
                seen.add(call_name)
        return result

    def setup_episodes(self, fold):
        """
        Parses into TodStructuredEpisode.
        """
        self.schemas = self.load_schemas()
        cache_path = os.path.join(self.dpath, f"{fold}_call_response_cache.json")

        if PathManager.exists(cache_path):
            with PathManager.open(cache_path, 'r') as f:
                self.call_response_cache = json.load(f)
            self.dbs = None
        else:
            self.call_response_cache = {}
            self.dbs = self.load_dbs()

        with PathManager.open(os.path.join(self.dpath, "dialog_acts.json")) as f:
            self.dialog_acts = json.load(f)

        chunks = self.load_chunks(fold)

        episodes = []
        for raw_episode in chunks:
            domains = raw_episode["services"]

            if self.opt.get("dialogue_id", "") != "":
                if raw_episode["dialogue_id"] != self.opt["dialogue_id"]:
                    continue

            skip = False  # need to skip outer for loop while in `for domains` inner for loop
            if self.opt.get("well_formatted_domains_only", True):
                if len(domains) == 0:
                    skip = True
                for domain in domains:
                    if domain not in WELL_FORMATTED_DOMAINS:
                        skip = True
            if skip:
                continue

            turn_id = 0  # matching naming in the `dialogues` files.
            turns = raw_episode["turns"]
            rounds = []
            goal_calls = []

            while turn_id < len(turns):
                goal, r = self._get_round(raw_episode['dialogue_id'], turns, turn_id)
                turn_id += 2
                rounds.append(r)

                if len(goal) > 0:
                    goal_calls.append(goal)

            episode = tod.TodStructuredEpisode(
                domain=tod.SerializationHelpers.inner_list_join(domains),
                api_schemas_machine=self._get_schemas_for_goal_calls(goal_calls),
                goal_calls_machine=goal_calls,
                rounds=rounds,
            )
            episodes.append(episode)

        with PathManager.open(cache_path, 'w') as f:
            json.dump(self.call_response_cache, f)

        return episodes

    def get_id_task_prefix(self):
        return "MultiwozV22"


class MultiWOZv22DSTTeacher(MultiwozV22Parser, tod_agents.TodUserSimulatorTeacher):
    """
    This Teacher is responsible for performing the task of Dialogue State Tracking.

    It can be used to evaluate LM on JGA (Joint Goal Accuracy) metric (as shown in
    [SimpleTOD](https://arxiv.org/abs/2005.00796) and
    [Soloist](https://arxiv.org/abs/2005.05298)).
    """

    BELIEF_STATE_DELIM = " ; "

    domains = [
        "attraction",
        "hotel",
        "hospital",
        "restaurant",
        "police",
        "taxi",
        "train",
    ]

    named_entity_slots = {
        "attraction--name",
        "restaurant--name",
        "hotel--name",
        "bus--departure",
        "bus--destination",
        "taxi--departure",
        "taxi--destination",
        "train--departure",
    }

    def __init__(self, opt: Opt, shared=None, *args, **kwargs):
        self.opt = opt
        self.fold = opt["datatype"].split(":")[0]
        opt["datafile"] = self.fold
        self.dpath = os.path.join(opt["datapath"], "multiwoz_v22")
        self.id = "multiwoz_v22"
        if self.opt.get("teacher_seed"):
            self.rng = np.random.RandomState(self.opt["teacher_seed"])
        else:
            self.rng = np.random.RandomState(SEED)

        if shared is None:
            build_.build(opt)
        super().__init__(opt, shared)

    def _load_data(self, fold):
        dataset_fold = "dev" if fold == "valid" else fold
        fold_path = os.path.join(self.dpath, dataset_fold)
        dialogs = []
        for file_id in range(1, fold_size(dataset_fold) + 1):
            filename = os.path.join(fold_path, f"dialogues_{file_id:03d}.json")
            with PathManager.open(filename, "r") as f:
                dialogs += json.load(f)
        return dialogs

    def _get_curr_belief_states(self, turn):
        belief_states = []
        for frame in turn["frames"]:
            if "state" in frame:
                if "slot_values" in frame["state"]:
                    for domain_slot_type in frame["state"]["slot_values"]:
                        for slot_value in frame["state"]["slot_values"][
                            domain_slot_type
                        ]:
                            domain, slot_type = domain_slot_type.split("-")
                            belief_state = f"{domain} {slot_type} {slot_value.lower()}"
                            belief_states.append(belief_state)
        return sorted(list(set(belief_states)))

    def _extract_slot_from_string(self, slots_string):
        """
        Either ground truth or generated result should be in the format: "dom slot_type
        slot_val, dom slot_type slot_val, ..., dom slot_type slot_val," and this
        function would reformat the string into list:

        ["dom--slot_type--slot_val", ... ]
        """

        slots_list = []
        per_domain_slot_lists = {}
        named_entity_slot_lists = []

        # split according to ";"
        str_split = slots_string.split(self.BELIEF_STATE_DELIM)

        if str_split[-1] == "":
            str_split = str_split[:-1]

        str_split = [slot.strip() for slot in str_split]

        for slot_ in str_split:
            slot = slot_.split()
            if len(slot) > 2 and slot[0] in self.domains:
                domain = slot[0]
                slot_type = slot[1]
                slot_val = " ".join(slot[2:])
                if not slot_val == "dontcare":
                    slots_list.append(domain + "--" + slot_type + "--" + slot_val)
                if domain in per_domain_slot_lists:
                    per_domain_slot_lists[domain].add(slot_type + "--" + slot_val)
                else:
                    per_domain_slot_lists[domain] = {slot_type + "--" + slot_val}
                if domain + "--" + slot_type in self.named_entity_slots:
                    named_entity_slot_lists.append(
                        domain + "--" + slot_type + "--" + slot_val
                    )
        return slots_list, per_domain_slot_lists, named_entity_slot_lists

    def custom_evaluation(
        self, teacher_action: Message, labels, model_response: Message
    ):
        """
        for dialog state tracking, we compute the joint goal accuracy, which is the
        percentage of the turns where the model correctly and precisely predicts all
        slots(domain, slot_type, slot_value).
        """
        resp = model_response.get("text")
        if not resp:
            return

        # extract ground truth from labels
        (
            slots_truth,
            slots_truth_per_domain,
            slots_truth_named_entity,
        ) = self._extract_slot_from_string(labels[0])

        # extract generated slots from model_response
        (
            slots_pred,
            slots_pred_per_domain,
            slots_pred_named_entity,
        ) = self._extract_slot_from_string(resp)

        for gt_slot in slots_truth:
            self.metrics.add("all/slot_r", AverageMetric(gt_slot in slots_pred))
            curr_domain = gt_slot.split("--")[0]
            self.metrics.add(
                f"{curr_domain}/slot_r", AverageMetric(gt_slot in slots_pred)
            )

        for gt_slot in slots_pred_named_entity:
            self.metrics.add(
                "hallucination", AverageMetric(gt_slot not in slots_truth_named_entity)
            )

        for predicted_slot in slots_pred:
            self.metrics.add("all/slot_p", AverageMetric(predicted_slot in slots_truth))
            curr_domain = predicted_slot.split("--")[0]
            self.metrics.add(
                f"{curr_domain}/slot_p", AverageMetric(predicted_slot in slots_truth)
            )

        self.metrics.add("jga", AverageMetric(set(slots_truth) == set(slots_pred)))
        self.metrics.add(
            "named_entities/jga",
            AverageMetric(
                set(slots_truth_named_entity) == set(slots_pred_named_entity)
            ),
        )
        for gt_slot in slots_truth_named_entity:
            self.metrics.add("all_ne/slot_r", AverageMetric(gt_slot in slots_pred))
            curr_domain = gt_slot.split("--")[0]
            self.metrics.add(
                f"{curr_domain}_ne/slot_r", AverageMetric(gt_slot in slots_pred)
            )
        for predicted_slot in slots_pred_named_entity:
            self.metrics.add(
                "all_ne/slot_p", AverageMetric(predicted_slot in slots_truth)
            )
            curr_domain = predicted_slot.split("--")[0]
            self.metrics.add(
                f"{curr_domain}_ne/slot_p", AverageMetric(predicted_slot in slots_truth)
            )

        for domain in slots_truth_per_domain:
            if domain in slots_pred_per_domain:
                self.metrics.add(
                    f"{domain}/jga",
                    AverageMetric(
                        slots_truth_per_domain[domain] == slots_pred_per_domain[domain]
                    ),
                )

    def setup_data(self, fold):
        dialogs = self._load_data(fold)
        examples = []
        for dialog in dialogs:
            context = []
            for turn in dialog["turns"]:
                curr_turn = turn["utterance"].lower()
                curr_speaker = (
                    "<user>" if turn["speaker"].lower() == "user" else "<system>"
                )
                curr_context = f"{curr_speaker} {curr_turn}"
                context.append(curr_context)
                cum_belief_states = self._get_curr_belief_states(turn)
                if curr_speaker == "<user>":
                    examples.append(
                        {
                            "dialogue_id": dialog["dialogue_id"],
                            "turn_num": turn["turn_id"],
                            "text": " ".join(context),
                            "labels": self.BELIEF_STATE_DELIM.join(cum_belief_states),
                        }
                    )
        if not is_this_circleci():
            self.rng.shuffle(examples)
        for example in examples:
            yield example, True


class UserSimulatorTeacher(MultiwozV22Parser, tod_agents.TodUserSimulatorTeacher):
    pass


class SystemTeacher(MultiwozV22Parser, tod_agents.TodSystemTeacher):
    pass


class StandaloneApiTeacher(MultiwozV22Parser, tod_agents.TodStandaloneApiTeacher):
    pass


class SingleGoalAgent(MultiwozV22Parser, tod_agents.TodSingleGoalAgent):
    pass


class SingleApiSchemaAgent(MultiwozV22Parser, tod_agents.TodSingleApiSchemaAgent):
    pass


class DefaultTeacher(SystemTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import copy
import json
from parlai.core.teachers import DialogTeacher
from .build import build


def _path(opt):
    # Build the data if it doesn't exist.
    build(opt)
    dt = opt['datatype'].split(':')[0]
    return os.path.join(
        opt['datapath'], 'reframe_thoughts', 'reframe_thoughts_dataset', dt + '.txt'
    )


class ReframeThoughtsTeacher(DialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['task'] = "reframe_thoughts"
        opt['datafile'] = _path(opt)
        super().__init__(opt, shared)

    def setup_data(self, datafile):
        all_data = [json.loads(line.strip()) for line in open(datafile)]
        for data in all_data:
            persona = data["persona"]
            pattern = data["pattern"] + " (" + data["pattern_def"] + ")"
            thought = data["thought"]
            persona_token = ""
            pattern_token = "[PAT]"
            thought_token = "[THT]"
            for reframe in data["reframes"]:
                input = " ".join(
                    [
                        persona_token,
                        persona,
                        pattern_token,
                        pattern,
                        thought_token,
                        thought,
                    ]
                ).strip()
                yield (input, reframe['reframe']), True


class DefaultTeacher(ReframeThoughtsTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from string import punctuation
import copy
from typing import List

from parlai.core.mutators import register_mutator, ManyEpisodeMutator, MessageMutator
from parlai.core.message import Message
from parlai.mutators.flatten import FlattenMutator
from parlai.utils.strings import normalize_reply as parlai_normalize_reply

from parlai.tasks.wizard_of_wikipedia.agents import TOKEN_KNOWLEDGE, TOKEN_END_KNOWLEDGE
from parlai.tasks.wizard_of_internet import constants as consts

from .constants import (
    FeedbackType,
    UNWANTED_TOKENS,
)


def normalize_reply(x):
    normalized_x = parlai_normalize_reply(x).strip()
    for t in UNWANTED_TOKENS:
        if normalized_x.lower().endswith(t):
            normalized_x = normalized_x[: -len(t)].strip()
    if '\n' in normalized_x:
        multilines = [
            ln.strip() for ln in normalized_x.split('\n') if len(ln.strip()) > 0
        ]
        if len(multilines) > 0:
            multilines = [
                (ln + ',' if ln[-1] not in punctuation else ln) for ln in multilines
            ]
            normalized_x = ' '.join(multilines)
            if normalized_x[-1] == ',':
                normalized_x = normalized_x[:-1] + '.'
    return normalized_x


def get_knowledge_sentence_with_special_tokens_only(
    knowledge_sentence: str, normalize_knowledge=False
):
    text = knowledge_sentence
    if normalize_knowledge:
        text = normalize_reply(text)
    return f"{TOKEN_KNOWLEDGE} {text} {TOKEN_END_KNOWLEDGE}"


def get_knowledge_appended_context(
    context: str, knowledge_sentence: str, normalize_knowledge=False
):
    return (
        context
        + "\n"
        + get_knowledge_sentence_with_special_tokens_only(
            knowledge_sentence, normalize_knowledge=normalize_knowledge
        )
    )


class FeedbackMixin(FlattenMutator):
    """
    Filter examples by last feedback, by subclassing FlattenMutator it is equivalent to.

    --mutators flatten+<this_mutator>
    """

    target_feedback: str

    def many_episode_mutation(self, episode):
        episodes = super().many_episode_mutation(episode)
        for episode in episodes:
            if episode[-1]['last_nonperfect_feedback'] == self.target_feedback:
                yield episode


@register_mutator("flatten_gold_human_mutator")
class GoldHuman(FeedbackMixin):
    """
    Filter examples by last nonperfect feedback.
    """

    target_feedback = FeedbackType.CORRECT_RESPONSE.value


@register_mutator("flatten_gold_bot_by_better_search_query_mutator")
class GoldBotByBetterSearchQuery(FeedbackMixin):
    """
    Filter examples by last nonperfect feedback.
    """

    target_feedback = FeedbackType.CORRECT_SEARCH_QUERY.value


@register_mutator("flatten_gold_bot_by_better_search_doc_mutator")
class GoldBotByBetterSearchDoc(FeedbackMixin):
    """
    Filter examples by last nonperfect feedback.
    """

    target_feedback = FeedbackType.CORRECT_DOC.value


@register_mutator("flatten_gold_bot_no_feedback_mutator")
class GoldBotNoFeedback(FeedbackMixin):
    """
    Filter examples by last nonperfect feedback.
    """

    target_feedback = FeedbackType.NONE.value


@register_mutator("flatten_gold_knowledge_response_mutator")
class GoldKnowledge(FlattenMutator):
    """
    give gold knowledge.
    """

    TOKEN_ENC_KNOWLEDGE = '__encknowledge__ '
    TOKEN_DEC_KNOWLEDGE = '__decknowledge__ '

    def many_episode_mutation(self, episode):
        episodes = super().many_episode_mutation(episode)
        for episode in episodes:
            assert len(episode) == 1
            if 'selected_sentences' in episode[-1]:
                new_episode = copy.deepcopy(episode)
                new_episode[-1].force_set(
                    'labels',
                    [
                        self.TOKEN_DEC_KNOWLEDGE
                        + ' '
                        + normalize_reply(episode[-1]['selected_sentences'])
                    ],
                )
                new_episode[-1].force_set(
                    'text', new_episode[-1]['text'] + ' ' + self.TOKEN_ENC_KNOWLEDGE
                )
                yield new_episode


@register_mutator("flatten_gold_knowledge_response_no_special_token_mutator")
class GoldKnowledgeNoSpecialToken(FlattenMutator):
    """
    give gold knowledge as target w/o special tokens.
    """

    def many_episode_mutation(self, episode):
        episodes = super().many_episode_mutation(episode)
        for episode in episodes:
            assert len(episode) == 1
            if 'selected_sentences' in episode[-1]:
                new_episode = copy.deepcopy(episode)
                new_episode[-1].force_set(
                    'labels',
                    [normalize_reply(episode[-1]['selected_sentences'])],
                )
                yield new_episode


@register_mutator("knowledge_appended_mutator")
class ContLnKnowledgeAppendedDialogueResponseTeacher(ManyEpisodeMutator):
    """
    parlai dd -t fits:mutators=flatten+knowledge_appended_mutator_internal appended gold
    knowleddge to response.
    """

    def many_episode_mutation(self, episode):
        out_episodes = []
        assert (
            len(episode) == 1
        ), 'knowledge_appended_mutator must be applied on flattened episode'
        for message in episode:
            if 'selected_sentences' in message:
                new_message = message.copy()
                new_message.force_set(
                    'text',
                    get_knowledge_appended_context(
                        new_message['text'],
                        new_message['selected_sentences'],
                        normalize_knowledge=True,
                    ),
                )
                out_episodes.append([new_message])
        return out_episodes


@register_mutator('no_woi_gold_docs_mutator')
class NoWoIGoldDocTeacher(MessageMutator):
    """
    Mutator that throw away WoI gold doc fields so the OBSERVATION_ECHO_RETRIEVER
    doesn't return top_docs.
    """

    def message_mutation(self, message: Message) -> Message:
        for key in [
            consts.RETRIEVED_DOCS_URLS,
            consts.RETRIEVED_DOCS_TITLES,
            consts.RETRIEVED_DOCS,
            'top_docs',
        ]:
            if key in message:
                del message[key]
        return message


@register_mutator('feedback_special_token_mutator')
class FeedbackSpecialToken(MessageMutator):
    """
    Mutator that throw away WoI gold doc fields so the OBSERVATION_ECHO_RETRIEVER
    doesn't return top_docs.
    """

    TOKEN_ENC_FEEDBACK = '__encfeedback__ '
    TOKEN_DEC_FEEDBACK = '__decfeedback__ '

    def _get_special_token_with_feedbacktypes(self, feedback_type):
        return self.TOKEN_DEC_FEEDBACK

    def message_mutation(self, message: Message) -> Message:
        message.force_set('text', message['text'] + ' ' + self.TOKEN_ENC_FEEDBACK)
        message.force_set(
            'labels',
            [
                self._get_special_token_with_feedbacktypes(
                    message['last_nonperfect_feedback']
                )
                + ' '
                + message['labels'][0]
            ],
        )
        return message


@register_mutator('feedback_type_specific_special_token_mutator')
class FeedbackTypeSpecificSpecialToken(FeedbackSpecialToken):
    """
    Mutators that add.
    """

    def _get_special_token_with_feedbacktypes(self, feedback_type):
        feedback_type = ''.join(feedback_type.split('_'))
        return f'__decfeedback{feedback_type}__'


@register_mutator('bot_failure_appended_mutator')
class BotFailureSpecialToken(MessageMutator):
    """
    Mutator that add the bot failure text in text (with special tokens surrounded)
    """

    TOKEN_BOT_FAILURE = '__botfailure__ '
    TOKEN_BOT_FAILURE_END = '__endbotfailure__ '

    def message_mutation(self, message: Message) -> Message:
        if 'bot_failure_text' in message:
            new_message = message.copy()
            new_message.force_set(
                'text',
                message['text']
                + f"\n{self.TOKEN_BOT_FAILURE} {normalize_reply(new_message['bot_failure_text'])} {self.TOKEN_BOT_FAILURE_END}",
            )
            return new_message

        return message


class SingleClassMixin(ManyEpisodeMutator):
    """
    Filter examples by class label.

    --mutators <this_mutator>
    """

    target_class: str

    def many_episode_mutation(self, episode):
        out_episodes = []
        for message in episode:
            if self.target_class in message['labels']:
                out_episodes.append([message])
        return out_episodes


@register_mutator("ok_mutator")
class OkClassFilter(SingleClassMixin):
    target_class = '__ok__'


@register_mutator("notok_mutator")
class NotOkClassFilter(SingleClassMixin):
    target_class = '__notok__'


@register_mutator('ok_pos_notok_neg_mutator')
class PosNegLabelConvertor(MessageMutator):
    """
    Mutator that throw away WoI gold doc fields so the OBSERVATION_ECHO_RETRIEVER
    doesn't return top_docs.
    """

    def message_mutation(self, message: Message) -> Message:
        new_message = copy.deepcopy(message)
        if message['labels'][0] == '__ok__':
            new_message.force_set('labels', ['pos'])
        elif message['labels'][0] == '__notok__':
            new_message.force_set('labels', ['neg'])
        return new_message


@register_mutator('fits_director_LTR_mutator')
class DirectorLeftToRightMutator(ManyEpisodeMutator):
    """
    DirectorLeftToRightMutator prepares data for training left to right (LTR) classifier
    for Director model.

    This limits to context to all but last utterance that is fed to the encoder.
    The final utterance is considered as a label for the decoder and the attribute/classifier
    labels are stored seperately marking the final utterance pos. or neg.

    This mutator also adds a is_ltr flag to differentiate classifier exs from the generator exs which are used to finetune the generator model.
    """

    def many_episode_mutation(self, episode: List[Message]) -> List[List[Message]]:
        new_episodes = []
        for message in episode:
            text = message['text']
            utterances = text.split('\n')

            if len(utterances) < 2:
                continue

            new_message = copy.deepcopy(message)
            new_message.force_set('is_ltr', True)
            if message['labels'][0] == 'pos':
                new_message.force_set('classifier_label', 'pos')
            else:
                new_message.force_set('classifier_label', 'neg')
            new_message.force_set('text', '\n'.join(utterances[:-1]))
            new_message.force_set('labels', [utterances[-1]])
            new_episodes.append([new_message])
        return new_episodes


@register_mutator('director_skip_reranking_mutator')
class DirectorSkipModififyingLogprobMutator(MessageMutator):
    """
    Mutator that adds a 'skip_director_reranking' key.
    """

    def message_mutation(self, message: Message) -> Message:
        message.force_set('skip_director_reranking', True)
        return message



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
import os
from parlai.utils.logging import logger
from parlai.core.build_data import DownloadableFile

VERSION = 'v0.1'


def _get_dpath_by_task_version(dpath, opt):
    if opt['fits_task_version'] == 'v1' and opt.get('unseen_task', False):
        return os.path.join(dpath, 'human_model_chats_unseen')

    if opt['fits_task_version'] == 'v1':
        return os.path.join(dpath, 'human_model_chats')
    if opt['fits_task_version'] == 'v2':
        return os.path.join(dpath, 'human_model_chats_v2')


RESOURCES = [
    DownloadableFile(
        f'http://parl.ai/downloads/fits/fits_data_{VERSION}.tar.gz',
        f'fits_data_{VERSION}.tar.gz',
        'fe85ab0b24dd904360cbe5b455e5eee502d5701eaca2b1be6d5b5f3fca19276e',
    )
]


def build(opt):
    version = VERSION
    # create particular instance of dataset depending on flags..
    dpath = os.path.join(opt['datapath'], 'fits')

    if not build_data.built(dpath, version):
        logger.warning('[build data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)
        # Mark the data as built.
        build_data.mark_done(dpath, version)

    return _get_dpath_by_task_version(dpath, opt)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import parlai.core.build_data as build_data
from parlai.core.build_data import DownloadableFile

import os

"""
eQASC Dataset. Multihop Question-Answering set with explanations.

EMNLP 2020 paper 'Learning to Explain: Datasets and Models for Identifying Valid Reasoning Chains in Multihop Question-Answering'
"""

RESOURCES = [
    DownloadableFile(
        url='1lllyoJLza2k5k8AlDVxJFHSl_pJRzFQb',
        file_name='eqasc_dev_grc.json',
        hashcode='515a34a92c4e69d54939d4b1c3c07fd17eb319d42b7cd0daf7fcfe81f372df8b',
        zipped=False,
        from_google=True,
    ),
    DownloadableFile(
        url='16jyHtTtSV_lb-RVZJZLi-HL8DI73XU0w',
        file_name='eqasc_test_grc.json',
        hashcode='524ecab20b0f635ab1d95d5ce437413c484c6280bfed5806e61c234f1f8c63b8',
        zipped=False,
        from_google=True,
    ),
    DownloadableFile(
        url='13AaSCgbRsqIxvmP-bBZpenvWaAQLs15G',
        file_name='eqasc_train_grc.json',
        hashcode='7603e5c908f4d0baac66f8288b2d6351eaa60d3ded2ea55f9f35b655dbbe0e26',
        zipped=False,
        from_google=True,
    ),
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'eqasc')
    version = '1.0'

    # check if data had been previously built
    if not build_data.built(dpath, version_string=version):
        print("[building data: " + dpath + "]")

        # make a clean directory if needed
        if build_data.built(dpath):
            # an older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # mark the data as built
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Base script for running TOD model-model chats.
"""

from collections import defaultdict
from copy import deepcopy
import json
import sys
from shutil import copyfile

from parlai.core.params import ParlaiParser
from parlai.core.script import ParlaiScript, register_script
from parlai.core.tod.tod_core import STANDARD_DONE
from parlai.utils.io import PathManager

from parlai.core.tod.tod_core import SerializationHelpers


@register_script("get_passing_only")
class GetPassingOnlyScript(ParlaiScript):
    @classmethod
    def setup_args(cls):
        parser = ParlaiParser(False, False)
        parser.add_argument("--cut-first-400", default=False)
        parser.add_argument("--convo-path", required=True)
        parser.add_argument(
            "--report-path",
            required=True,
            help="path of the report saved from the tod_metrics_script",
        )
        parser.add_argument(
            "--print-to-file",
            default=False,
            help="save the results to a file (by hackishly redirecting stdout)",
        )
        parser.add_argument(
            "--filter-call-attempts",
            default=True,
            help="when True, only counts as 'passing' if System made exactly the same # of api calls as # of goals",
        )
        return parser

    def run(self):
        opt = self.opt

        with PathManager.open(opt["report_path"]) as r:
            print(opt["report_path"])
            report = json.load(r)["report"]
        tod_metrics = report["tod_metrics"]

        if opt["cut_first_400"]:
            tod_metrics = tod_metrics[400:]

        infile_base = opt["convo_path"].replace(".jsonl", "")
        outfile_base = infile_base + "_processed"

        copyfile(f"{infile_base}.metadata", f"{outfile_base}.metadata")

        if opt["print_to_file"]:
            print_file_path = infile_base.replace("_conversations", "_processed_stats")
            print_file = open(print_file_path, "w+")
            orig_stdout = sys.stdout
            sys.stdout = print_file

        found = defaultdict(lambda: 0)
        not_found = defaultdict(lambda: 0)

        with PathManager.open(f"{outfile_base}.jsonl", "w") as out:
            with PathManager.open(opt["convo_path"]) as c:
                lines = c.readlines()
                if opt["cut_first_400"]:
                    lines = lines[400:]
                print("len lines: ", len(lines))
                print("len tod_metrics: ", len(tod_metrics))
                for i, l in enumerate(lines):
                    if i >= len(tod_metrics):
                        break
                    goals = SerializationHelpers.str_to_goals(
                        tod_metrics[i]["goal"]["text"][len(STANDARD_DONE) :]
                    )
                    if len(goals) == 0:
                        print(goals)
                        continue
                    if tod_metrics[i]["synthetic_task_success"] == 1.0 and (
                        not self.opt["filter_call_attempts"]
                        or tod_metrics[i].get("api_call_attempts", 1)  # legacy
                        == len(goals)
                    ):
                        print(goals, tod_metrics[i]["api_call_attempts"], l)
                        self.api_string_add(found, goals, 1)
                        self.api_string_add(not_found, goals, 0)
                        out.write(l)
                    else:
                        self.api_string_add(not_found, goals, 1)
                        self.api_string_add(found, goals, 0)
            print("count found", sum(found.values()))
            print("count notfound", sum(not_found.values()))
            print(
                "============================ FOUND ======================\n",
                [
                    (k, v)
                    for k, v in sorted(found.items(), key=lambda x: x[1], reverse=True)
                ],
            )
            print(
                "============================= NOT FOUND ====================\n",
                [
                    (k, v)
                    for k, v in sorted(
                        not_found.items(), key=lambda x: x[1], reverse=True
                    )
                ],
            )

        biggest_misses = deepcopy(not_found)
        for f in found:
            biggest_misses[f] -= found[f]
        print(
            "======================= BIGGEST MISSES (# not found - # found) ======================\n",
            "\n".join(
                [
                    json.dumps((k, v))
                    for k, v in sorted(
                        biggest_misses.items(), key=lambda x: x[1], reverse=True
                    )
                ]
            ),
        )
        fraction = {}
        for k in not_found:
            total = not_found[k] + found[k]
            fraction[k] = (float(not_found[k]) / total, total)
        print(
            "=========================== BIGGEST FRACTIONAL DELTAS (not_found / total # )  ====================\n",
            "\n".join(
                [
                    json.dumps((k, v))
                    for k, v in sorted(
                        fraction.items(), key=lambda x: x[1], reverse=True
                    )
                ]
            ),
        )

        if opt["print_to_file"]:
            print_file.close()
            sys.stdout = orig_stdout

    def api_string_add(self, found_list, api_strings, val):
        for api_string in api_strings:
            if "api_name" not in api_string:
                continue
            api_name = api_string["api_name"]
            api_string["api_name:" + api_name] = ""
            api_name = api_string.pop("api_name")
            found_list[json.dumps(sorted(list(api_string)))] += val
            api_string["api_name"] = api_name


if __name__ == "__main__":
    GetPassingOnlyScript.main()



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Download and build the data if it does not exist.

import os

import parlai.core.build_data as build_data
from parlai.core.build_data import DownloadableFile

RESOURCES = [
    DownloadableFile(
        'https://drive.google.com/u/1/uc?id=1d-987AjyfVdFnohBqQ3yaHt-b5mVezHa&export=download&confirm=t',
        'prosocial_dialog_v1.tar.gz',
        '112e402e283949cbc36b67a86877c8aea098a7fe40fd3e180095e1e147958eba',
    )
]


def build(opt):
    dpath = os.path.join(opt['datapath'], 'prosocial_dialog')
    version = '0.1'

    if not build_data.built(dpath, version_string=version):
        print('[building data: ' + dpath + ']')
        if build_data.built(dpath):
            # An older version exists, so remove these outdated files.
            build_data.remove_dir(dpath)
        build_data.make_dir(dpath)

        # Download the data.
        for downloadable_file in RESOURCES:
            downloadable_file.download_file(dpath)

        # Mark the data as built.
        build_data.mark_done(dpath, version_string=version)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# Copyright (c) Microsoft Corporation.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# Code adapted from https://github.com/microsoft/fastseq/blob/main/fastseq/ops/ngram_repeat_block.py.

"""
Wrapper for ngram_repeat_block cuda extension.
"""
import parlai.utils.logging as logging
import torch
from torch import nn

import os
from torch.utils.cpp_extension import load


class NGramRepeatBlock(nn.Module):
    """
    Wrapper class for calling ngram_repeat_block cuda extension.
    """

    def __init__(self):
        super(NGramRepeatBlock, self).__init__()
        current = os.getcwd()
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

        try:
            self.ngram_repeat_block_cuda = load(
                name='ngram_repeat_block_cuda',
                sources=[
                    '../clib/cuda/ngram_repeat_block_cuda.cpp',
                    '../clib/cuda/ngram_repeat_block_cuda_kernel.cu',
                ],
            )
        except Exception as e:
            logging.warning(f"Unable to load ngram blocking on GPU: {e}")
            self.ngram_repeat_block_cuda = None

        os.chdir(current)

    def forward(
        self,
        hypothesis,
        context,
        lprobs,
        bsz,
        step,
        beam_size,
        no_repeat_ngram_size,
        if_context_blocking=False,
    ):
        """
        Args:
        hypothesis(Tensor): (beam*bsz, current_sequence_length)
        context(Tensor): context for context-blocking
        lprobs(Tensor): likelihood probability(beam, vocab_size)
        bsz(int): batch size
        step(int): current step
        beam_size(int): beam size
        no_repeat_ngram_size(int): Ngram size
        if_context_blocking(bool): whether to use context-blocking
        """
        # placeholder tensor to pass in to pass type check, won't be used
        if not if_context_blocking:
            context = torch.Tensor([0]).long()
        assert hypothesis.size(0) == bsz * beam_size
        assert lprobs.size(0) == bsz * beam_size
        hypothesis = hypothesis.contiguous()
        context = context.contiguous()
        lprobs = lprobs.contiguous()
        return self.ngram_repeat_block_cuda.forward(
            hypothesis,
            context,
            lprobs,
            bsz,
            step,
            beam_size,
            no_repeat_ngram_size,
            if_context_blocking,
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import random
from typing import List, Tuple

from parlai.core.opt import Opt
import parlai.tasks.convai2.agents as convai2
import parlai.tasks.blended_skill_talk.agents as bst
import parlai.tasks.empathetic_dialogues.agents as ed
import parlai.tasks.msc.agents as msc
import parlai.tasks.wizard_of_wikipedia.agents as wow
import parlai.utils.logging as logging

import projects.bb3.constants as BB3_CONST
from projects.bb3.tasks.module_level_tasks import BB3TeacherMixin

from projects.seeker.tasks.search_decision import (
    WoiSearchDecisionTeacher as woi_sdm,
    SquadSearchDecisionTeacher as squad_sdm,
    TriviaQASearchDecisionTeacher as triviaqa_sdm,
    NQSearchDecisionTeacher as nq_sdm,
    Convai2SearchDecisionTeacher as convai2_sdm,
    EDSearchDecisionTeacher as ed_sdm,
    MSCSearchDecisionTeacher as msc_sdm,
)


class WoiSearchDecisionTeacher(BB3TeacherMixin, woi_sdm):
    pass


class WowSearchDecisionTeacher(BB3TeacherMixin, wow.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten',
                'wow_maybe_generate_search_query_mutator',
                'skip_retrieval_mutator',
                'filter_wow_topic_only_search_decision_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['add_missing_turns'] = 'all'
        super().__init__(opt, shared)


class SquadSearchDecisionTeacher(BB3TeacherMixin, squad_sdm):
    pass


class TriviaQASearchDecisionTeacher(BB3TeacherMixin, triviaqa_sdm):
    pass


class NQOpenSearchDecisionTeacher(BB3TeacherMixin, nq_sdm):
    pass


class Convai2SearchDecisionTeacher(BB3TeacherMixin, convai2_sdm):
    pass


class EDSearchDecisionTeacher(BB3TeacherMixin, ed_sdm):
    pass


class MSCSearchDecisionTeacher(BB3TeacherMixin, msc_sdm):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'flatten',
                'filter_silence_only_mutator',
                'skip_retrieval_mutator',
                'bst_tasks_maybe_generate_search_query_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['include_session1'] = False
        super().__init__(opt, shared)


def get_memory_decision_task_mutator(opt: Opt) -> str:
    """
    Set the mutators appropriately for the memory decision tasks.
    """
    mutators = '+'.join(
        [
            'memory_decision_mutator',
            'skip_retrieval_mutator',
            'prompt_memory_decision_mutator',
        ]
    )
    if opt.get('mutators'):
        mutators = '+'.join([mutators, opt['mutators']])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


def balance_data(episodes: List[Tuple]) -> List[Tuple]:
    """
    Balance the data so that we have equal numbers of memory/no memory.

    Very hacky and requires deep understanding of DialogData objects.

    Should in theory be used in a script to dump to disk.
    """
    random.seed(42)
    pos_inds = set()
    neg_inds = set()
    for i, ep in enumerate(episodes):
        # assume flattened
        assert len(ep) == 1
        ex = ep[0]
        label = ex[1][0] if isinstance(ex[1], list) else ex[1]
        if label == BB3_CONST.DO_ACCESS_MEMORY:
            pos_inds.add(i)
        elif label == BB3_CONST.DONT_ACCESS_MEMORY:
            neg_inds.add(i)
        else:
            raise ValueError(f"Incorrect label: {label}")
    pos_exs = [e for i, e in enumerate(episodes) if i in pos_inds]
    neg_exs = [e for i, e in enumerate(episodes) if i in neg_inds]
    new_data = []
    num_to_sample = min(len(pos_exs), len(neg_exs))
    if num_to_sample == len(pos_exs):
        new_data = pos_exs + random.sample(neg_exs, num_to_sample)
    else:
        new_data = neg_exs + random.sample(pos_exs, num_to_sample)

    return new_data


class Convai2MemoryDecisionTeacher(BB3TeacherMixin, convai2.NormalizedTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_decision_task_mutator(opt)
        opt['task'] += ':no_cands'
        super().__init__(opt, shared)


class EDMemoryDecisionTeacher(BB3TeacherMixin, ed.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_decision_task_mutator(opt)
        super().__init__(opt, shared)


class BSTMemoryDecisionTeacher(BB3TeacherMixin, bst.DefaultTeacher):
    def __init__(self, opt, shared=None):
        opt['mutators'] = get_memory_decision_task_mutator(opt)
        super().__init__(opt, shared)


class MSCMemoryDecisionTeacher(BB3TeacherMixin, msc.DefaultTeacher):
    def __init__(self, opt, shared=None):
        mutators = '+'.join(
            [
                'memory_decision_mutator',
                'skip_retrieval_mutator',
                'prompt_memory_decision_mutator',
                'filter_silence_only_memory_decision_mutator',
            ]
        )
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        opt['include_session1'] = False
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import random


"""
Functions to generate safe curated response that repalce an Agent's main response.
"""

####################################################################################
#
#   Init prompt: initial message to start a chit-chat.

STARTING_PROMPT_MESSAGES = (
    "Hey I am here. Let's chat!",
    "Howdy!",
    "Hey. What's up?",
    "Hello there!",
    "Hey, how are you doing today?",
    "Hello, how is your day going so far?",
    "Hi, are you into sports?",
    "Have you seen any interesting movies recently?",
    "Hi. I can't wait to chat with you!",
    "Hi! What's new with you?",
    "Hello, it's a pleasure to meet you.",
)


def generate_init_prompot():
    return random.choice(STARTING_PROMPT_MESSAGES)


####################################################################################
#
# Safe response
#
# A fall back method to avoid discussing sensetive and/or inappropriate topics.
# This may be used to generate a safe response that substitues the main response from an agent,
# in case the main response was flagged as inappropriate.

# The generated responses consist of three parts.
# 1-AVOID: Avoiding to continue the conversation about that topic, for example, "I don't know"
# 2-TEMPLATE: a template to start talking about a new topic, for example, " have you heard about "
# 3-TOPIC: the next topic to steer the conversation towards, for example, "new marvel movie"


AVOID = (
    "I don't know anything about that, sorry.",
    "I am not sure I know about that.",
    "Sorry! I'm confused.",
    "Uh, not sure what you meant.",
    "Whoops, that's not something I know anything about.",
    "I do not know what you mean when you say that.",
    "Huh? I don't get it.",
    "Umm? I don't understand what that means.",
)


TEMPLATE = [
    ("What do you think of", "?"),
    ("Can you help me understand", "?"),
    ("What if we discuss", "?"),
    ("You know what though? It would help me a lot if you could educate me about", "."),
    ("Why don't we talk about", "?"),
    ("You know what I wish I knew more about ---", "."),
    ("Instead, how about you teach me about", "?"),
    ("Could you enlighten me with your thoughts on", "?"),
    ("Let's have a chat about", "."),
    ("It could be interesting to have a conversation on", "."),
    ("I want to hear your opinion on", "."),
    ("Ooh you know what would be fun? Let's discuss", "."),
    ("I recently read an interesting article about", "."),
    ("Guess what? I just learned something new about", "."),
    ("Actually, can we just shoot the breeze about", "."),
    ("How about we discuss", "?"),
    ("Why don't have a conversation about", "?"),
    ("I would rather talk about", "."),
    ("What I really want to talk about is", "."),
    ("Maybe it is better if we discuss", "."),
    ("Can we talk instead about", "?"),
]

TOPIC = (
    "the weather we are getting next weekend",
    "traveling to exotic destinations",
    "the world's best vacation destinations",
    "reading new books",
    "space travel",
    "the possibility of life on Mars",
    "astronomy",
    "how the universe was formed",
    "the moon",
    "the ocean",
    "philosophy",
    "fashion",
    "volcanoes",
    "geology",
    "climate change",
    "artificial intelligence",
    "celebrities",
    "classical music",
    "rock and roll bands",
    "hip-hop music",
    "country music",
    "modern art",
    "money",
    "taking risks",
    "non-fiction books",
    "fiction novels",
    "ancient Egypt",
    "the Ancient Romans",
    "Greek Mythology",
    "medieval castles",
    "dragons",
    "yo-yos",
    "dominoes",
    "blanket and pillow forts",
    "basketball players",
    "football",
    "soccer stars",
    "hockey teams",
    "tennis doubles partners",
    "the economy",
    "broken supply chains",
    "panda bears",
    "current events",
    "middle school cliques",
    "high school sweethearts",
    "the best new phone apps",
    "what the metaverse might be like",
    "giraffes",
    "penguins",
    "walruses",
    "American presidents",
    "Antarctica",
    "the Bermuda Triangle",
    "falling in love",
    "online dating",
    "best friends",
    "poisonous plants",
    "gothic architecture",
    "prehistoric sea creatures",
    "Area 51",
    "robot vacuums",
    "self-driving cars",
    "meditation",
    "how to relax in stressful situations",
    "waking up before it is light out in the morning",
    "talent shows",
    "secret societies",
    "holiday baking",
    "birthday party themes",
    "the best spots to celebrate New Year's Eve",
    "geometry",
    "algebra",
    "New York City",
    "Silicon Valley",
    "Los Angeles",
    "Tokyo",
    "Paris",
    "London",
    "the Arctic Circle",
    "Siberia",
    "the Gobi Desert",
    "the Himalayas",
    "the Andes Mountains",
    "quinoa",
    "kale",
    "green smoothies",
    "peanut allergies",
    "english literature",
    "pie",
    "bagels and the best types of schmear",
    "cake",
    "cookies",
    "ice cream",
    "pastries",
    "breakfast cereal",
    "reality tv",
    "elephants",
    "polar bears",
    "good investment strategies",
    "zoos",
    "dinosaurs",
    "archeologists",
    "kittens",
    "puppies",
    "the Scottish Highlands",
    "the Loch Ness monster",
    "aliens",
    "giant squids",
    "charities",
    "things that make people happy",
    "things that interrupt a good night's sleep",
    "cryptocurrency",
    "the inevitable heat death of our Universe",
    "unusual hobbies",
    "superheroes",
    "home improvement tv shows",
    "hidden super-powers",
)


def generate_safe_response():
    avoid = random.choice(AVOID)
    template, punctuation = random.choice(TEMPLATE)
    topic = random.choice(TOPIC)
    sentence = " ".join([avoid, template, topic]) + punctuation
    return sentence



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import os
from .build import build
from parlai.core.teachers import FbDeprecatedDialogTeacher

'''This dataset is available in seven different languages.
To use the dataset in the specified language, use the task flag to specify it.

--task xpersona:{LANGUAGE}

The default language is English.
'''


def _path(opt):
    # build the data if it does not exist
    build(opt)

    # set up path to data (specific to each dataset)
    dt = opt['datatype'].split(':')[0]
    if ':' in opt['task']:
        language = opt['task'].split(':')[1]
    else:
        language = 'En'
    return os.path.join(opt['datapath'], 'XPersona', language + '_' + dt + '.txt')


class XPersonaTeacher(FbDeprecatedDialogTeacher):
    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['datafile'] = _path(opt)
        super().__init__(opt, shared)


class DefaultTeacher(XPersonaTeacher):
    pass


class EnTeacher(XPersonaTeacher):
    pass


class ItTeacher(XPersonaTeacher):
    pass


class ZhTeacher(XPersonaTeacher):
    pass


class IdTeacher(XPersonaTeacher):
    pass


class FrTeacher(XPersonaTeacher):
    pass


class JpTeacher(XPersonaTeacher):
    pass


class KoTeacher(XPersonaTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import copy
import json
import random
from typing import Optional

from parlai.utils.misc import warn_once
from parlai.utils.data import DatatypeHelper
from parlai.core.opt import Opt
from parlai.core.teachers import DialogTeacher
from parlai.core.params import ParlaiParser
from parlai.core.message import Message
from parlai.utils.io import PathManager


from parlai.tasks.wizard_of_internet import constants as woi_consts
from .mutators import (
    normalize_reply,
    get_knowledge_appended_context,
    get_knowledge_sentence_with_special_tokens_only,
)
from .build import build
from .constants import (
    FeedbackType,
    TITLE_PASSAGE_DELIMITER,
    OK_LABEL,
    NOT_OK_LABEL,
)


def format_to_woi_consistent(docs, woi_format_gold_docs):
    retrieved_docs = []
    retrieved_doc_titles = []
    for doc_str in docs:
        doc_splits = doc_str.split(TITLE_PASSAGE_DELIMITER)
        doc_passage = TITLE_PASSAGE_DELIMITER.join(doc_splits[1:])
        retrieved_docs.append(doc_passage)
        retrieved_doc_titles.append(doc_splits[0])

    woi_format_gold_docs.update(
        {
            woi_consts.RETRIEVED_DOCS: retrieved_docs,
            woi_consts.RETRIEVED_DOCS_TITLES: retrieved_doc_titles,
            woi_consts.RETRIEVED_DOCS_URLS: ['' for _ in retrieved_docs],
        }
    )
    return woi_format_gold_docs


def dedup_list(arr):
    new_arr = []
    for e in arr:
        if e not in new_arr:
            new_arr.append(e)
    return new_arr


class FitsBaseTeacher(DialogTeacher):
    """
    Base Teacher that extract metadata for dialogues.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('Fits Base options')
        agent.add_argument(
            '--metrics-per-feedback',
            type=str,
            default=None,
            help='metrics to show when breaking down by last feedback that takes effect; if None then do not break down by feedbacktype',
        )
        agent.add_argument(
            '--add-bot-failure-to-feedback-input',
            type=bool,
            default=False,
            help='If True, then include bot failure turns in the context, otherwise don"t ',
        )
        agent.add_argument(
            '--add-knowledge-sentence-to-feedback-input',
            type=bool,
            default=False,
            help='If True, then include knowledge sentence in the context, otherwise don"t ',
        )
        agent.add_argument(
            '--fits-task-version',
            type=str,
            default='v1',
            help='dataset version, v1 = 7.5k dialogues, v2 = 13.5k dialogues',
        )
        agent.add_argument(
            '--unseen-task',
            type=bool,
            default=False,
            help='dataset skills seen by default and unseen if set True',
        )
        return parser

    def __init__(self, opt, shared=None):
        opt['datafile'] = build(opt)
        self.is_training = DatatypeHelper.is_training(opt['datatype'])
        self.metrics_per_feedback = opt.get('metrics_per_feedback', None)
        self.add_bot_failure_to_feedback_input = opt.get(
            'add_bot_failure_to_feedback_input', False
        )
        self.add_knowledge_sentence_to_feedback_input = opt.get(
            'add_knowledge_sentence_to_feedback_input', False
        )
        self.unseen_task = opt.get('unseen_task', False)
        super().__init__(opt, shared)
        self.id = f'fits_{opt["mutators"]}_{opt["fits_task_version"]}_seen{opt["unseen_task"]}'

    def _load_raw_data(self, data_path):
        print('loading: ' + data_path)

        if self.datatype.startswith('train'):
            path_to_open = os.path.join(data_path, 'train.txt')
        elif self.datatype.startswith('valid'):
            path_to_open = os.path.join(data_path, 'valid.txt')
        else:
            path_to_open = os.path.join(data_path, 'test.txt')

        if self.unseen_task:
            assert self.datatype.startswith('test')
            path_to_open = os.path.join(data_path, 'test.txt')
            warn_once('No train/valid splits for unseen tasks')

        with PathManager.open(path_to_open) as f:
            raw_data = [json.loads(line.strip()) for line in f]

        return raw_data

    def setup_data(self, data_path):
        raw_data = self._load_raw_data(data_path)
        data = []
        for chat in raw_data:
            perfect_episode, _ = self._get_perfect_episode(
                chat, is_training=self.is_training
            )
            data.append(perfect_episode)

        for episode in data:
            start_idx = 0
            for i, turn in enumerate(episode):
                yield Message(turn), i == start_idx

    def _add_feedback_to_perfect_utt(
        self, bot_acts, human_acts, perfect_dialog_context
    ):
        feedback_chain = [
            bot_act.get('improve_bucket', FeedbackType.PERFECT.value)
            for bot_act in bot_acts
        ]
        feedback_examples = []
        effective_human_feedback = {}
        human_revision = {'search_query': [], 'selected_sentences': []}
        bot_gold = {'search_query': []}
        bot_no_feedback = {'search_query': []}
        bot_modular_failure = {'search_query': []}
        woi_format_gold_docs = {
            woi_consts.SELECTED_SENTENCES: []
        }  # otherwise get_retrieved_knowledge will break
        if 'search_queries' in bot_acts[0]:
            woi_format_gold_docs = format_to_woi_consistent(
                bot_acts[0]['top_docs'], woi_format_gold_docs
            )

        if len(bot_acts) == 1:
            # no feedback turn
            if 'search_queries' in bot_acts[0]:
                bot_no_feedback['search_query'].append(bot_acts[0]['search_queries'])
        else:
            for i, bot_act in enumerate(bot_acts):
                # skip updating the top_docs or search_queries if the bot response is from human better_response
                if (
                    i > 0
                    and bot_acts[i - 1].get('improve_bucket', '')
                    != FeedbackType.CORRECT_RESPONSE.value
                    and 'top_docs' in bot_act
                    and isinstance(bot_act['top_docs'], list)
                ):
                    woi_format_gold_docs = format_to_woi_consistent(
                        bot_act['top_docs'], woi_format_gold_docs
                    )
                # clear the knowledge sentence selection if it's from a fresh search
                if i > 0 and bot_acts[i - 1].get('improve_bucket', '') in [
                    FeedbackType.CORRECT_SEARCH_QUERY.value
                ]:
                    woi_format_gold_docs[woi_consts.SELECTED_SENTENCES] = []

                if (
                    bot_act.get('improve_bucket', '')
                    == FeedbackType.CORRECT_SEARCH_QUERY.value
                ):
                    effective_human_feedback = {
                        'search_query': bot_acts[i + 1]['search_queries']
                    }
                    human_revision['search_query'].append(
                        effective_human_feedback['search_query']
                    )
                    if 'search_queries' in bot_act:
                        bot_modular_failure['search_query'].append(
                            bot_act['search_queries']
                        )
                elif (
                    bot_act.get('improve_bucket', '') == FeedbackType.CORRECT_DOC.value
                ):
                    if 'better_response' in effective_human_feedback:
                        del effective_human_feedback['better_response']
                    effective_human_feedback['selected_sentences'] = human_acts[i + 1][
                        'feedback_text'
                    ]
                    human_revision['selected_sentences'].append(
                        effective_human_feedback['selected_sentences']
                    )
                    woi_format_gold_docs[woi_consts.SELECTED_SENTENCES] = [
                        effective_human_feedback['selected_sentences']
                    ]
                    bot_gold['search_query'].append(bot_acts[i]['search_queries'])
                elif (
                    bot_act.get('improve_bucket', '')
                    == FeedbackType.CORRECT_RESPONSE.value
                ):
                    effective_human_feedback['better_response'] = human_acts[i + 1][
                        'feedback_text'
                    ]

                # collect feedback examples as target
                if bot_act.get('improve_bucket', '') in [
                    FeedbackType.CORRECT_SEARCH_QUERY.value,
                    FeedbackType.CORRECT_DOC.value,
                    FeedbackType.CORRECT_RESPONSE.value,
                ]:
                    assert (
                        human_acts[i + 1]['feedback_type'] == bot_act['improve_bucket']
                    )
                    new_context_texts = [normalize_reply(human_acts[0]['text'])]
                    if self.add_bot_failure_to_feedback_input:
                        if (
                            self.add_knowledge_sentence_to_feedback_input
                            and woi_consts.SELECTED_SENTENCES in woi_format_gold_docs
                            and woi_format_gold_docs[woi_consts.SELECTED_SENTENCES]
                        ):
                            new_context_texts.append(
                                get_knowledge_sentence_with_special_tokens_only(
                                    woi_format_gold_docs[woi_consts.SELECTED_SENTENCES][
                                        -1
                                    ],
                                    normalize_knowledge=True,
                                )
                            )
                        new_context_texts.append(normalize_reply(bot_act['text']))

                    feedback_examples.append(
                        {
                            'text': '\n'.join(
                                perfect_dialog_context + [t for t in new_context_texts]
                            ),
                            'labels': [normalize_reply(human_acts[i + 1]['text'])],
                            'last_nonperfect_feedback': bot_act['improve_bucket'],
                            'bot_failure_text': bot_act['text'],
                            **(copy.deepcopy(woi_format_gold_docs)),
                        }
                    )
                    if 'search_queries' in bot_act:
                        feedback_examples[-1]['search_queries'] = bot_act[
                            'search_queries'
                        ]
                    if 'selected_sentences' in effective_human_feedback:
                        feedback_examples[-1][
                            'selected_sentences'
                        ] = effective_human_feedback['selected_sentences']

        effective_human_feedback['feedback_chain'] = feedback_chain
        effective_human_feedback['last_nonperfect_feedback'] = (
            feedback_chain[-2] if len(bot_acts) > 1 else FeedbackType.NONE.value
        )

        # remove redundant examples
        human_revision['search_query'] = dedup_list(human_revision['search_query'])
        bot_gold['search_query'] = dedup_list(
            [
                q
                for q in bot_gold['search_query']
                if q not in human_revision['search_query']
            ]
        )
        human_revision['selected_sentences'] = dedup_list(
            human_revision['selected_sentences']
        )

        perfect_bot_utt_with_feedback = {
            'text': bot_acts[-1]['text'],
            'human_revision': human_revision,
            'bot_gold': bot_gold,
            'bot_no_feedback': bot_no_feedback,
            'bot_modular_failure': bot_modular_failure,
            'human_acts': human_acts,
            'bot_acts': bot_acts,
        }
        for k, v in effective_human_feedback.items():
            if v is not None:
                perfect_bot_utt_with_feedback[k] = v

        if woi_format_gold_docs:
            for k, v in woi_format_gold_docs.items():
                if k in perfect_bot_utt_with_feedback:
                    raise RuntimeError("name collision in continual learning task")
                perfect_bot_utt_with_feedback[k] = v

        return [
            {'text': human_acts[0]['text']},
            perfect_bot_utt_with_feedback,
        ], feedback_examples

    def _get_perfect_episode(self, chat, is_training):
        raw_dialog = chat['dialog']
        assert len(raw_dialog) % 2 == 0, "odd number of utterances"
        perfect_dialog = []
        # store all bot/human acts before the bot generate the next perfect response
        bot_acts = []
        human_acts = []
        perfect_dialog_context = []
        raw_feedback_examples = []
        for i, utt in enumerate(raw_dialog):
            # whether it's human message or bot message
            if utt.get('id') in ['YOU', 'Speaker 1']:
                human_acts.append(utt)
            else:
                if (
                    i == len(raw_dialog) - 1
                    and utt.get('improve_bucket', FeedbackType.PERFECT.value)
                    != FeedbackType.PERFECT.value
                ):
                    utt['improve_bucket'] = FeedbackType.PERFECT.value
                bot_acts.append(utt)
                if (
                    utt.get('improve_bucket', FeedbackType.PERFECT.value)
                    == FeedbackType.PERFECT.value
                    or i == len(raw_dialog) - 1
                ):
                    (
                        utts_with_feedbacks,
                        feedback_examples,
                    ) = self._add_feedback_to_perfect_utt(
                        bot_acts, human_acts, perfect_dialog_context
                    )
                    perfect_dialog.extend(utts_with_feedbacks)
                    perfect_dialog_context.extend(
                        [normalize_reply(t['text']) for t in utts_with_feedbacks]
                    )
                    raw_feedback_examples.extend(feedback_examples)
                    bot_acts = []
                    human_acts = []

        perfect_episode = []
        chat_metadata = {
            'specific_task': chat['chat_task_name'],
            'task_requirement': chat['chat_task_requirement'],
            'topic': chat['topic'],
            'generic_topic': chat['generic_topic'],
            'domain': chat['domain'],
            'task_uid': chat['chat_task_uid'],
        }
        for i in range(0, len(perfect_dialog) - 1, 2):
            action = {
                'id': self.id,
                'text': normalize_reply(perfect_dialog[i]['text']),
                'labels': [normalize_reply(perfect_dialog[i + 1]['text'])],
            }
            for k, v in perfect_dialog[i + 1].items():
                if k not in ['text', 'id', 'labels']:
                    action[k] = v
            action.update(chat_metadata)
            perfect_episode.append(action)

        feedback_examples_with_metadata = []
        for act in raw_feedback_examples:
            feedback_examples_with_metadata.append(
                {
                    **act,
                    **chat_metadata,
                }
            )

        return perfect_episode, feedback_examples_with_metadata

    def observe(self, observation):
        """
        Process observation for metrics.
        """
        obs = super().observe(observation)
        if self.last_act is None:
            return obs
        has_update = False
        if (
            self.metrics_per_feedback is not None
            and self.last_act.get('last_nonperfect_feedback', None) is not None
        ):
            for key, value in observation['metrics'].items():
                if (
                    self.metrics_per_feedback == 'all'
                    or key in self.metrics_per_feedback
                ):
                    self.metrics.add(
                        f"{self.last_act['last_nonperfect_feedback']}_{key}", value
                    )
                    has_update = True

        if has_update:
            updated_recent_metrics = self.metrics.report_recent()
            if updated_recent_metrics and 'metrics' in obs:
                obs.pop('metrics')
                obs['metrics'] = updated_recent_metrics
        return obs


class FitsFeedbackTeacher(FitsBaseTeacher):
    """
    Teacher that display free-form text feedbacks.
    """

    def __init__(self, opt, shared=None):
        super().__init__(opt, shared)
        self.id = f'fits_feedback_{opt["mutators"]}_botTurn{self.add_bot_failure_to_feedback_input}'

    def setup_data(self, data_path):
        raw_data = self._load_raw_data(data_path)

        data = []
        for chat in raw_data:
            _, feedback_examples_with_metadata = self._get_perfect_episode(
                chat, is_training=self.is_training
            )
            data.extend(feedback_examples_with_metadata)

        for turn in data:
            message = Message(turn)
            yield message, True


class FitsSatisfactionTeacher(FitsBaseTeacher):
    """
    Teacher that display binary classes __ok__, __notok__ on bot response.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group(
            'Continual Learning New Tasks Satisfaction Teacher options'
        )
        agent.add_argument(
            '--prepend-task-specifics',
            type=bool,
            default=False,
            help='whether to prepend task name and task requirement to the dialogue context',
        )
        agent.add_argument(
            '--must-have-knowledge-sentence',
            type=bool,
            default=False,
            help='whether to only include examples with perfect/gold knowledge sentence',
        )
        agent.add_argument(
            '--positive-example-source',
            type=str,
            default='all',
            choices=['human_gold', 'human_and_bot_gold_revision', 'all'],
            help='source of positive examples concatenated by "," ',
        )
        return parser

    def __init__(self, opt, shared=None):
        opt_cp = copy.deepcopy(opt)
        opt_cp['add_bot_failure_to_feedback_input'] = True
        self.must_have_knowledge_sentence = opt_cp['must_have_knowledge_sentence']
        if self.must_have_knowledge_sentence:
            opt_cp['add_knowledge_sentence_to_feedback_input'] = True
        if (
            not self.must_have_knowledge_sentence
            and opt_cp['add_knowledge_sentence_to_feedback_input']
        ):
            warn_once(
                "You are loading Satisfaction Teacher with a mixture of examples both w/ and w/o knowledge"
            )
        self.prepend_task_specifics = opt_cp['prepend_task_specifics']
        self.positive_example_source = opt_cp.get('positive_example_source', 'all')
        super().__init__(opt_cp, shared)
        self.id = f'fits_satisfaction_{opt_cp["mutators"]}_prependtask_{self.prepend_task_specifics}_{self.positive_example_source}'

    def setup_data(self, data_path):
        raw_data = self._load_raw_data(data_path)

        data = []
        neg_data = []
        pos_data = []

        def _get_prepend_task_info(ex):
            return (
                '__task__ '
                + normalize_reply(ex['specific_task'])
                + ' | '
                + normalize_reply(ex['task_requirement'])
                + '__endtask__'
            )

        for chat in raw_data:
            (
                perfect_episode,
                feedback_examples_with_metadata,
            ) = self._get_perfect_episode(chat, is_training=self.is_training)
            previous_texts = []
            for ex in feedback_examples_with_metadata:
                if self.must_have_knowledge_sentence:
                    if not (
                        woi_consts.SELECTED_SENTENCES in ex
                        and ex[woi_consts.SELECTED_SENTENCES]
                    ):
                        continue
                text = ex['text']
                if self.prepend_task_specifics:
                    task_context = _get_prepend_task_info(ex)
                    if task_context:
                        text = task_context + '\n' + text
                # dedup sample
                if text in previous_texts:
                    continue
                previous_texts.append(text)
                new_exs = copy.deepcopy(ex)
                new_exs.update(
                    {
                        'text': text,
                        'labels': [NOT_OK_LABEL],
                        'last_nonperfect_feedback': ex['last_nonperfect_feedback'],
                    }
                )
                neg_data.append(new_exs)
            history = []
            for turn in perfect_episode:
                history.append(turn['text'])
                history.append(turn['labels'][0])
                text = '\n'.join(history)
                if self.must_have_knowledge_sentence:
                    if not (
                        woi_consts.SELECTED_SENTENCES in turn
                        and turn[woi_consts.SELECTED_SENTENCES]
                    ):
                        continue
                    else:
                        text = get_knowledge_appended_context(
                            '\n'.join(history[:-1]),
                            turn[woi_consts.SELECTED_SENTENCES][-1],
                            normalize_knowledge=True,
                        )
                        text = '\n'.join([text] + history[-1:])
                if self.prepend_task_specifics:
                    task_context = _get_prepend_task_info(turn)
                    if task_context:
                        text = task_context + '\n' + text
                if (
                    self.positive_example_source == 'human_gold'
                    and turn['last_nonperfect_feedback']
                    != FeedbackType.CORRECT_RESPONSE.value
                ) or (
                    self.positive_example_source == 'human_and_bot_gold_revision'
                    and turn['last_nonperfect_feedback'] == FeedbackType.NONE.value
                ):
                    continue
                new_exs = copy.deepcopy(turn)
                new_exs.update(
                    {
                        'text': text,
                        'labels': [OK_LABEL],
                        'last_nonperfect_feedback': FeedbackType.NONE.value,
                    }
                )
                pos_data.append(new_exs)

        data = neg_data + pos_data
        random.shuffle(data)
        for turn in data:
            message = Message(turn)
            yield message, True


class QueryTeacher(FitsBaseTeacher):
    """
    Query for finetune query generator.

    human_gold: human written search query that lead to perfect utterance,
        e.g better_search_query->perfect; better_search_query -> better_search_doc -> perfect;
            better_search_query -> better_search_doc -> better_response -> perfect
    human_bronze: human written search query that doesn't belong to perfect utterance
    bot_gold: bot search queries that leads to perfect utterance e.g. better_doc -> perfect
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group(
            'Continual Learning New Tasks Query Teacher options'
        )
        agent.add_argument(
            '--query-source',
            type=str,
            default='human_gold',
            choices=[
                'human_gold',
                'human_bronze',
                'bot_gold',
                'human',
                'all',
            ],
            help='metrics to show when breaking down by bot nickname',
        )
        return parser

    def __init__(self, opt, shared=None):
        self.query_source = opt.get('query_source', None)
        super().__init__(opt, shared)
        self.id = f'fits_query_{opt["mutators"]}'

    def setup_data(self, data_path):
        raw_data = self._load_raw_data(data_path)
        data = []
        for chat in raw_data:
            perfect_episode, _ = self._get_perfect_episode(
                chat, is_training=self.is_training
            )
            feedbacks = self._extract_feedback(perfect_episode)
            if self.query_source in ['human_gold', 'human', 'all']:
                data.extend(feedbacks['human_gold'])
            if self.query_source in ['human_bronze', 'human', 'all']:
                data.extend(feedbacks['human_bronze'])
            if self.query_source in ['bot_gold', 'all']:
                data.extend(feedbacks['bot_gold'])

        for turn in data:
            yield Message(turn), True

    def _extract_feedback(self, episode):
        history = []
        res = {
            'human_gold': [],
            'human_bronze': [],
            'bot_gold': [],
            'bot_failure': [],
        }

        for turn in episode:
            history.append(normalize_reply(turn['text']))
            human_provided_queries = turn['human_revision']['search_query']
            per_turn_res = {
                'human_gold': [],
                'human_bronze': [],
                'bot_gold': [],
                'bot_failure': [],
            }
            if (
                turn['last_nonperfect_feedback']
                == FeedbackType.CORRECT_SEARCH_QUERY.value
                or (
                    # ..., better_search_query, better_search_doc, perfect
                    len(turn['feedback_chain']) >= 3
                    and turn['feedback_chain'][-3]
                    == FeedbackType.CORRECT_SEARCH_QUERY.value
                    and turn['feedback_chain'][-2] == FeedbackType.CORRECT_DOC.value
                )
                or (
                    # ..., better_search_query, better_search_doc, better_others, perfect
                    len(turn['feedback_chain']) >= 4
                    and turn['feedback_chain'][-4]
                    == FeedbackType.CORRECT_SEARCH_QUERY.value
                    and turn['feedback_chain'][-3] == FeedbackType.CORRECT_DOC.value
                    and turn['feedback_chain'][-2]
                    == FeedbackType.CORRECT_RESPONSE.value
                )
            ):
                per_turn_res['human_gold'] = [human_provided_queries[-1]]
            per_turn_res['human_bronze'] = [
                q for q in human_provided_queries if q not in per_turn_res['human_gold']
            ]
            per_turn_res['bot_gold'] = turn['bot_gold'].get('search_query', [])
            per_turn_res['bot_failure'] = turn['bot_modular_failure'].get(
                'search_query', []
            )

            for source in per_turn_res:
                for q in per_turn_res[source]:
                    res[source].append(
                        {'text': '\n'.join(history), 'labels': [normalize_reply(q)]}
                    )
            history.append(normalize_reply(turn['labels'][0]))
        return res


class DefaultTeacher(FitsBaseTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from typing import Optional

from parlai.tasks.reasoning.base import (
    AbstractReasoningTeacher,
    AbstractReason,
    AbstractQuestionAnswer,
)

from parlai.tasks.reasoning.question_answer import (
    SimpleQuestionAnswer,
    PremiseHypothesisQuestionAnswer,
)

from parlai.tasks.reasoning.reason_types.step_by_step import (
    StepByStepReason,
    MWPStepsReason,
)
from parlai.tasks.reasoning.reason_types.free_form import FreeFormReason


class StepByStepReasoningTeacher(AbstractReasoningTeacher):
    """
    Downstream Teachers should implement:

    * get_data_for_fold(self, fold) which yields dicts that contain at least:
        * "steps" (to work with StepByStepReason)
        * "question" (to work with SimpleQuestionAnswer)
        * "answer" (to work with SimpleQuestionAnswer)
    """

    @classmethod
    def get_reason_class(self) -> AbstractReason:
        return StepByStepReason

    @classmethod
    def get_question_answer_class(self) -> AbstractQuestionAnswer:
        return SimpleQuestionAnswer


class MWPStepsReasoningTeacher(AbstractReasoningTeacher):
    """
    Downstream Teachers should implement:

    * get_data_for_fold(self, fold) which yields dicts that contain at least:
        * "steps" (to work with MWPStepsReason)
        * "question" (to work with SimpleQuestionAnswer)
        * "answer" (to work with SimpleQuestionAnswer)
    """

    @classmethod
    def get_reason_class(self) -> AbstractReason:
        return MWPStepsReason

    @classmethod
    def get_question_answer_class(self) -> AbstractQuestionAnswer:
        return SimpleQuestionAnswer


class NliTeacher(AbstractReasoningTeacher):
    """
    Downstream Teachers should implement:

    * get_data_for_fold(self, fold) which yields dicts that contain at least:
        * "reason" (to work with FreeFormReason)
        * "premise" (to work with PremiseHypothesisQuestionAnswer)
        * "hypotehsis" (to work with PremiseHypothesisQuestionAnswer)
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.set_defaults(
            task_prompt="Given the premise, is the hypothesis an entailment, a contradiction, or neutral?"
        )
        return parser

    @classmethod
    def get_reason_class(self) -> AbstractReason:
        return FreeFormReason

    @classmethod
    def get_question_answer_class(self) -> AbstractQuestionAnswer:
        return PremiseHypothesisQuestionAnswer



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
Wrapper object holding metrics for TODWorld.

This class is in its own file to prevent circular dependencies + monolithic files.
"""

from parlai.core.message import Message
from parlai.core.metrics import Metrics
from parlai.core.tod.tod_core import (
    TodAgentType,
    TOD_AGENT_TYPE_TO_PREFIX,
    SerializationHelpers,
    STANDARD_GOAL,
)
from typing import Any, Dict
import parlai.core.tod.world_metrics_handlers as world_metrics_handlers

# Change the following to define which Metrics Handlers are used in TodWorld.
# The ones used below are from `world_metrics_handlers.py` only. However, See `parlai/projects/tod_simulator/world_metrics/extended_world_metrics.py` for others.

WORLD_METRIC_HANDLERS = [
    world_metrics_handlers.AllGoalApiCallSuccessMetricsHandler,
    world_metrics_handlers.UserGeneratedDoneMetricHandler,
]


class TodMetrics(Metrics):
    """
    Helper container which encapsulates TOD metrics and does some basic prepocessing to
    handlers to calculate said metrics.

    This class should generally not need to be changed; add new metrics handlers to
    `WORLD_METRIC_HANDLERS` (or otherwise override `self.handlers` of this class) to
    change metrics actively being used.
    """

    def __init__(self, shared: Dict[str, Any] = None) -> None:
        super().__init__(shared=shared)
        self.handlers = [x() for x in WORLD_METRIC_HANDLERS]
        self.convo_started = False
        self.last_episode_metrics = Metrics()

    def handle_message(self, message: Message, agent_type: TodAgentType):
        if "text" not in message:
            return
        if agent_type == TodAgentType.GOAL_GROUNDING_AGENT and len(
            message["text"]
        ) > len(STANDARD_GOAL):
            # Only count a conversation as started if there is a goal.
            self.convo_started = True
        for handler in self.handlers:
            metrics = self._handle_message_impl(message, agent_type, handler)
            if metrics is not None:
                for name, metric in metrics.items():
                    if metric is not None:
                        self.add(name, metric)

    def _handle_message_impl(
        self,
        message: Message,
        agent_type: TodAgentType,
        handler: world_metrics_handlers.TodMetricsHandler,
    ):
        prefix_stripped_text = message["text"].replace(
            TOD_AGENT_TYPE_TO_PREFIX[agent_type], ""
        )
        if agent_type is TodAgentType.API_SCHEMA_GROUNDING_AGENT:
            return handler.handle_api_schemas(
                message, SerializationHelpers.str_to_api_schemas(prefix_stripped_text)
            )
        if agent_type is TodAgentType.GOAL_GROUNDING_AGENT:
            return handler.handle_goals(
                message, SerializationHelpers.str_to_goals(prefix_stripped_text)
            )
        if agent_type is TodAgentType.USER_UTT_AGENT:
            return handler.handle_user_utt(message, prefix_stripped_text)
        if agent_type is TodAgentType.API_CALL_AGENT:
            return handler.handle_api_call(
                message, SerializationHelpers.str_to_api_dict(prefix_stripped_text)
            )
        if agent_type is TodAgentType.API_RESP_AGENT:
            return handler.handle_api_resp(
                message, SerializationHelpers.str_to_api_dict(prefix_stripped_text)
            )
        if agent_type is TodAgentType.SYSTEM_UTT_AGENT:
            return handler.handle_sys_utt(message, prefix_stripped_text)

    def get_last_episode_metrics(self):
        """
        This is a bit of a hack so that we can  report whether or not a convo has
        successfully hit all goals and associate this with each episode for the purposes
        of doing filtering.
        """
        return self.last_episode_metrics

    def episode_reset(self):
        self.last_episode_metrics = None
        if self.convo_started:
            self.last_episode_metrics = Metrics()
            for handler in self.handlers:
                metrics = handler.get_episode_metrics()
                handler.episode_reset()
                if metrics is not None:
                    for name, metric in metrics.items():
                        if metric is not None:
                            self.add(name, metric)
                            self.last_episode_metrics.add(name, metric)
            self.convo_started = False



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.utils.testing import AutoTeacherTest


class TestDefaultTeacher(AutoTeacherTest):
    task = 'light_multiparty'


class TestSpeakerPredictionTeacher(AutoTeacherTest):
    task = f'light_multiparty:SpeakerPredictionTeacher'


class TestSpeakersTeacher(AutoTeacherTest):
    task = f'light_multiparty:SpeakerPredictionTeacher'


class TestFirstSpeakerTeacher(AutoTeacherTest):
    task = f'light_multiparty:FirstSpeakerTeacher'


class TestSecondSpeakerTeacher(AutoTeacherTest):
    task = f'light_multiparty:SecondSpeakerTeacher'


class TestThirdSpeakerTeacher(AutoTeacherTest):
    task = f'light_multiparty:ThirdSpeakerTeacher'
